from .utils import get_default_workload_template, kwargs_from_env, load_template
from .websocket import WebSocketAPI

__all__ = [
    "kwargs_from_env",
    "get_default_workload_template",
    "load_template",
    "WebSocketAPI",
]



"""
Custom exceptions for the CPLN Python client.
"""

from typing import Any, Dict, Optional


class CPLNError(Exception):
    """
    Base exception class for CPLN client errors.

    This exception is raised when there's an error interacting with the CPLN API
    or when there's an error in the client itself.
    """

    def __init__(
        self,
        message: str,
        status_code: Optional[int] = None,
        response: Optional[Dict[str, Any]] = None,
        request_id: Optional[str] = None,
    ) -> None:
        """
        Initialize a new CPLNError.

        Args:
            message: A human-readable error message
            status_code: HTTP status code if the error came from an API call
            response: The response body if the error came from an API call
            request_id: The request ID if available from the API
        """
        self.message = message
        self.status_code = status_code
        self.response = response
        self.request_id = request_id

        # Construct the full error message
        error_parts = [message]
        if status_code:
            error_parts.append(f"Status code: {status_code}")
        if request_id:
            error_parts.append(f"Request ID: {request_id}")
        if response:
            error_parts.append(f"Response: {response}")

        super().__init__("\n".join(error_parts))


class AuthenticationError(CPLNError):
    """Raised when there's an authentication error with the CPLN API."""

    def __init__(
        self,
        message: str = "Authentication failed",
        status_code: Optional[int] = None,
        response: Optional[Dict[str, Any]] = None,
        request_id: Optional[str] = None,
    ) -> None:
        super().__init__(message, status_code, response, request_id)


class ValidationError(CPLNError):
    """Raised when there's a validation error in the request parameters."""

    def __init__(
        self,
        message: str = "Validation error",
        status_code: Optional[int] = None,
        response: Optional[Dict[str, Any]] = None,
        request_id: Optional[str] = None,
    ) -> None:
        super().__init__(message, status_code, response, request_id)


class ResourceNotFoundError(CPLNError):
    """Raised when a requested resource is not found."""

    def __init__(
        self,
        message: str = "Resource not found",
        status_code: Optional[int] = None,
        response: Optional[Dict[str, Any]] = None,
        request_id: Optional[str] = None,
    ) -> None:
        super().__init__(message, status_code, response, request_id)


class RateLimitError(CPLNError):
    """Raised when the API rate limit is exceeded."""

    def __init__(
        self,
        message: str = "Rate limit exceeded",
        status_code: Optional[int] = None,
        response: Optional[Dict[str, Any]] = None,
        request_id: Optional[str] = None,
    ) -> None:
        super().__init__(message, status_code, response, request_id)



class GVCApiMixin:
    """
    A mixin class that provides GVC (Global Virtual Cluster) related API methods.
    """

    def get_gvc(self, name: str = None):
        """
        Retrieves GVC information from the Control Plane API.

        Args:
            name (str, optional): The name of the specific GVC to retrieve.
                If not provided, returns a list of all GVCs.

        Returns:
            dict: The GVC information or list of GVCs

        Raises:
            APIError: If the request fails
        """
        endpoint = "gvc"
        if name:
            endpoint += f"/{name}"
        return self._get(endpoint)

    def create_gvc(self, name: str, description: str = None):
        """
        Creates a new GVC in the Control Plane.

        Args:
            name (str): The name of the GVC to create
            description (str, optional): A description of the GVC

        Raises:
            ValueError: Currently not implemented
        """
        raise ValueError(
            "Not implemented! The payload to do this is annoyingly long, so somebody else do it."
        )

    def delete_gvc(self, name: str):
        """
        Deletes a GVC from the Control Plane.

        Args:
            name (str): The name of the GVC to delete

        Returns:
            requests.Response: The response from the API

        Raises:
            APIError: If the request fails
        """
        return self._delete(f"gvc/{name}")



from typing import Any, Dict, Optional

import requests

from ..errors import APIError, NotFound
from .config import APIConfig
from .gvc import GVCApiMixin
from .image import ImageApiMixin
from .workload import (
    WorkloadApiMixin,
    WorkloadDeploymentMixin,
)


class APIClient(
    requests.Session,
    GVCApiMixin,
    ImageApiMixin,
    WorkloadApiMixin,
    WorkloadDeploymentMixin,
):
    """
    A low-level client for the Control Plane API.

    Example:

        >>> import cpln
        >>> client = cpln.APIClient(
            base_url='https://api.cpln.io/'
        )
        >>> client.version()

    Args:
        base_url (str): URL to the Control Plane server.
        version (str): The version of the API to use. Set to ``auto`` to
            automatically detect the server's version. Default: ``1.0.0``
        timeout (int): Default timeout for API calls, in seconds.
    """

    def __init__(self, config: Optional[APIConfig] = None, **kwargs):
        super().__init__()

        # Initialize the config object
        if config is None:
            config = APIConfig(**kwargs)

        self.config = config

    def _get(self, endpoint: str) -> Dict[str, Any]:
        """
        Makes a GET request to the specified API endpoint.

        Args:
            endpoint (str): The API endpoint to request

        Returns:
            dict: The JSON response from the API

        Raises:
            NotFound: If the resource is not found
            APIError: If the API returns an error
        """
        resp = self.get(f"{self.config.org_url}/{endpoint}", headers=self._headers)

        # Handle error responses
        if resp.status_code == 404:
            raise NotFound(f"Resource not found: {endpoint}")
        elif resp.status_code >= 400:
            raise APIError(f"API error ({resp.status_code}): {resp.text}")

        return resp.json()

    def _delete(self, endpoint: str) -> requests.Response:
        """
        Makes a DELETE request to the specified API endpoint.

        Args:
            endpoint (str): The API endpoint to delete

        Returns:
            requests.Response: The response object from the API

        Raises:
            NotFound: If the resource is not found
            APIError: If the API returns an error
        """
        resp = self.delete(f"{self.config.org_url}/{endpoint}", headers=self._headers)

        # Handle error responses
        if resp.status_code == 404:
            raise NotFound(f"Resource not found: {endpoint}")
        elif resp.status_code >= 400:
            raise APIError(f"API error ({resp.status_code}): {resp.text}")

        return resp

    def _post(
        self, endpoint: str, data: Optional[Dict[str, Any]] = None
    ) -> requests.Response:
        """
        Makes a POST request to the specified API endpoint.

        Args:
            endpoint (str): The API endpoint to post to
            data (dict, optional): The data to send in the request body

        Returns:
            requests.Response: The response object from the API

        Raises:
            APIError: If the API returns an error
        """
        resp = self.post(
            f"{self.config.org_url}/{endpoint}",
            json=data,
            headers=self._headers,
        )

        # Handle error responses
        if resp.status_code >= 400:
            error_msg = f"API error ({resp.status_code})"
            try:
                error_data = resp.json()
                if isinstance(error_data, dict) and "error" in error_data:
                    error_msg = f"{error_msg}: {error_data['error']}"
                else:
                    error_msg = f"{error_msg}: {resp.text}"
            except (ValueError, KeyError):
                error_msg = f"{error_msg}: {resp.text}"

            raise APIError(error_msg)

        return resp

    def _patch(
        self, endpoint: str, data: Optional[Dict[str, Any]] = None
    ) -> requests.Response:
        """
        Makes a PATCH request to the specified API endpoint.

        Args:
            endpoint (str): The API endpoint to update
            data (dict, optional): The data to send in the request body

        Returns:
            requests.Response: The response object from the API

        Raises:
            NotFound: If the resource is not found
            APIError: If the API returns an error
        """
        resp = self.patch(
            f"{self.config.org_url}/{endpoint}",
            json=data,
            headers=self._headers,
        )

        # Handle error responses
        if resp.status_code == 404:
            raise NotFound(f"Resource not found: {endpoint}")
        elif resp.status_code >= 400:
            raise APIError(f"API error ({resp.status_code}): {resp.text}")

        return resp

    @property
    def _headers(self):
        return {"Authorization": f"Bearer {self.config.token}"}



from enum import Enum
from typing import Dict, Optional


class PostgresExitCode(Enum):
    """Enumeration of PostgreSQL command exit codes and their meanings.
    This class handles exit codes for both pg_dump and pg_restore commands."""

    SUCCESS = 0
    GENERAL_ERROR = 1
    INVALID_ARGUMENTS = 2
    CONNECTION_ERROR = 3
    PERMISSION_DENIED = 4
    FILE_ERROR = 5
    PROCESS_INTERRUPTED = 130
    UNKNOWN_ERROR = 255

    @classmethod
    def get_message(cls, code: int, command: str = "PostgreSQL command") -> str:
        """Get the human-readable message for an exit code.

        Args:
            code (int): The exit code to get the message for
            command (str): The name of the PostgreSQL command (e.g., 'pg_dump', 'pg_restore')
        """
        messages: Dict[int, str] = {
            cls.SUCCESS.value: f"The {command} completed successfully.",
            cls.GENERAL_ERROR.value: f"An error occurred while executing {command}. This could be due to invalid command-line arguments, internal errors, or other issues.",
            cls.INVALID_ARGUMENTS.value: f"Invalid command-line arguments were provided to {command}.",
            cls.CONNECTION_ERROR.value: "Failed to connect to the database server. Check your connection parameters and ensure the server is running.",
            cls.PERMISSION_DENIED.value: f"Permission denied while accessing the database or {command} files.",
            cls.FILE_ERROR.value: f"Error accessing files for {command}. Check file permissions and available disk space.",
            cls.PROCESS_INTERRUPTED.value: f"The {command} process was interrupted by a signal (e.g., Ctrl+C).",
            cls.UNKNOWN_ERROR.value: f"An unknown error occurred while executing {command}.",
        }
        return messages.get(code, f"Unknown {command} exit code: {code}")

    @classmethod
    def is_error(cls, code: int) -> bool:
        """Check if an exit code indicates an error."""
        return code != cls.SUCCESS.value

    @classmethod
    def get_error_type(cls, code: int) -> Optional[str]:
        """Get the type of error for an exit code."""
        error_types: Dict[int, str] = {
            cls.GENERAL_ERROR.value: "General Error",
            cls.INVALID_ARGUMENTS.value: "Invalid Arguments",
            cls.CONNECTION_ERROR.value: "Connection Error",
            cls.PERMISSION_DENIED.value: "Permission Denied",
            cls.FILE_ERROR.value: "File Error",
            cls.PROCESS_INTERRUPTED.value: "Process Interrupted",
            cls.UNKNOWN_ERROR.value: "Unknown Error",
        }
        return error_types.get(code)



import unittest
from typing import Any, Dict, List, Tuple, cast
from unittest.mock import MagicMock, Mock, patch

from cpln.config import WorkloadConfig
from cpln.errors import WebSocketExitCodeError
from cpln.models.workloads import Workload, WorkloadCollection
from requests import Response


class TestWorkload(unittest.TestCase):
    """Tests for the Workload class"""

    def setUp(self) -> None:
        """Set up the test"""
        self.attrs: Dict[str, Any] = {
            "id": "test-workload-id",
            "name": "test-workload",
            "description": "Test workload description",
            "spec": {
                "containers": [{"name": "app", "image": "nginx:latest"}],
                "defaultOptions": {
                    "suspend": "false",
                    "autoscaling": {
                        "metric": "memory",
                        "minReplicas": 1,
                        "maxReplicas": 5,
                    },
                    "capacityAI": True,
                },
                "type": "standard",
            },
        }
        self.client: MagicMock = MagicMock()
        self.collection: MagicMock = MagicMock()
        self.state: Dict[str, str] = {"gvc": "test-gvc"}
        self.workload = Workload(
            attrs=self.attrs,
            client=self.client,
            collection=self.collection,
            state=self.state,
        )

    def test_initialization(self) -> None:
        """Test workload initialization"""
        self.assertEqual(self.workload.attrs, self.attrs)
        self.assertEqual(self.workload.client, self.client)
        self.assertEqual(self.workload.collection, self.collection)
        self.assertEqual(self.workload.state, self.state)
        # The string representation includes the class name and workload name
        self.assertTrue("Workload" in str(self.workload))
        self.assertTrue("test-workload" in str(self.workload))

    def test_get(self) -> None:
        """Test get method"""
        expected_response: Dict[str, Any] = {"name": "test-workload", "updated": True}
        self.client.api.get_workload.return_value = expected_response

        result = self.workload.get()

        self.assertEqual(result, expected_response)
        self.client.api.get_workload.assert_called_once_with(self.workload.config())

    def test_delete(self) -> None:
        """Test delete method"""
        # Mock print to avoid output during test
        with patch("builtins.print"):
            self.workload.delete()

        self.client.api.delete_workload.assert_called_once_with(self.workload.config())

    def test_suspend(self) -> None:
        """Test suspend method"""
        # Mock print to avoid output during test
        with patch("builtins.print"):
            self.workload.suspend()

        self.client.api.patch_workload.assert_called_once_with(
            config=self.workload.config(),
            data={"spec": {"defaultOptions": {"suspend": "true"}}},
        )

    def test_unsuspend(self) -> None:
        """Test unsuspend method"""
        # Mock print to avoid output during test
        with patch("builtins.print"):
            self.workload.unsuspend()

        self.client.api.patch_workload.assert_called_once_with(
            config=self.workload.config(),
            data={"spec": {"defaultOptions": {"suspend": "false"}}},
        )

    def test_exec_success(self) -> None:
        """Test exec method with success"""
        command: str = "echo 'Hello, World!'"
        location: str = "test-location"
        expected_response: Dict[str, str] = {"output": "Hello, World!"}
        self.client.api.exec_workload.return_value = expected_response

        result = self.workload.exec(command, location)

        self.assertEqual(result, expected_response)
        self.client.api.exec_workload.assert_called_once_with(
            config=self.workload.config(location=location), command=command
        )

    def test_exec_error(self) -> None:
        """Test exec method with error"""
        command: str = "invalid-command"
        location: str = "test-location"

        # Create a WebSocketExitCodeError with exit_code
        error = WebSocketExitCodeError("Command failed")
        error.exit_code = 1

        self.client.api.exec_workload.side_effect = error

        # Mock print to avoid output during test
        with (
            patch("builtins.print"),
            self.assertRaises(WebSocketExitCodeError),
        ):
            self.workload.exec(command, location)

        self.client.api.exec_workload.assert_called_once_with(
            config=self.workload.config(location=location), command=command
        )

    def test_ping_success(self) -> None:
        """Test ping method with success"""
        location: str = "test-location"
        self.client.api.exec_workload.return_value = {"output": "ping"}

        result = self.workload.ping(location)

        self.assertEqual(result["status"], 200)
        self.assertEqual(result["message"], "Successfully pinged workload")
        self.assertEqual(result["exit_code"], 0)
        self.client.api.exec_workload.assert_called_once()

    def test_ping_websocket_error(self) -> None:
        """Test ping method with WebSocketExitCodeError"""
        location: str = "test-location"

        # Create a WebSocketExitCodeError with exit_code
        error = WebSocketExitCodeError("Command failed")
        error.exit_code = 1

        self.client.api.exec_workload.side_effect = error

        result = self.workload.ping(location)

        self.assertEqual(result["status"], 500)
        self.assertIn("Command failed with exit code", result["message"])
        self.assertEqual(result["exit_code"], 1)

    def test_ping_general_exception(self) -> None:
        """Test ping method with general exception"""
        location: str = "test-location"
        self.client.api.exec_workload.side_effect = Exception("Connection failed")

        result = self.workload.ping(location)

        self.assertEqual(result["status"], 500)
        self.assertEqual(result["message"], "Connection failed")
        self.assertEqual(result["exit_code"], -1)

    def test_export(self) -> None:
        """Test export method"""
        # Mock the get method to avoid API call
        with patch.object(self.workload, "get"):
            result = self.workload.export()

            self.assertEqual(result["name"], self.attrs["name"])
            self.assertEqual(result["gvc"], self.state["gvc"])
            self.assertEqual(result["spec"], self.attrs["spec"])

    def test_config_default(self) -> None:
        """Test config method with default parameters"""
        config = self.workload.config()

        self.assertIsInstance(config, WorkloadConfig)
        self.assertEqual(config.gvc, self.state["gvc"])
        self.assertEqual(config.workload_id, self.attrs["name"])
        self.assertIsNone(config.location)

    def test_config_with_location(self) -> None:
        """Test config method with location parameter"""
        location: str = "test-location"
        config = self.workload.config(location=location)

        self.assertEqual(config.location, location)

    def test_config_with_custom_gvc(self) -> None:
        """Test config method with custom gvc parameter"""
        custom_gvc: str = "custom-gvc"
        config = self.workload.config(gvc=custom_gvc)

        self.assertEqual(config.gvc, custom_gvc)

    def test_get_remote(self) -> None:
        """Test get_remote method"""
        location: str = "test-location"
        expected_remote: str = "https://test-remote"
        self.client.api.get_remote.return_value = expected_remote

        result = self.workload.get_remote(location)

        self.assertEqual(result, expected_remote)
        self.client.api.get_remote.assert_called_once_with(
            self.workload.config(location=location)
        )

    def test_get_remote_wss(self) -> None:
        """Test get_remote_wss method"""
        location: str = "test-location"
        expected_remote: str = "wss://test-remote"
        self.client.api.get_remote_wss.return_value = expected_remote

        result = self.workload.get_remote_wss(location)

        self.assertEqual(result, expected_remote)
        self.client.api.get_remote_wss.assert_called_once_with(
            self.workload.config(location=location)
        )

    def test_get_replicas(self) -> None:
        """Test get_replicas method"""
        location: str = "test-location"
        expected_replicas: List[str] = ["replica1", "replica2"]
        self.client.api.get_replicas.return_value = expected_replicas

        result = self.workload.get_replicas(location)

        self.assertEqual(result, expected_replicas)
        self.client.api.get_replicas.assert_called_once_with(
            self.workload.config(location=location)
        )

    def test_get_containers(self) -> None:
        """Test get_containers method"""
        location: str = "test-location"
        expected_containers: List[str] = ["container1", "container2"]
        self.client.api.get_containers.return_value = expected_containers

        result = self.workload.get_containers(location)

        self.assertEqual(result, expected_containers)
        self.client.api.get_containers.assert_called_once_with(
            self.workload.config(location=location)
        )

    def test_clone_basic(self) -> None:
        """Test clone method with basic parameters"""
        # Setup test data
        new_name: str = "cloned-workload"

        # Mock successful API response
        mock_response: Mock = Mock(spec=Response)
        mock_response.status_code = 201
        mock_response.text = "Created"
        self.client.api.create_workload.return_value = mock_response

        # Mock the export method to return the expected data
        with patch.object(self.workload, "export") as mock_export:
            mock_export.return_value = {
                "name": self.attrs["name"],
                "gvc": self.state["gvc"],
                "spec": self.attrs["spec"],
            }

            # Mock print to avoid output during test
            with patch("builtins.print"):
                # Call clone method - should not raise any exceptions
                self.workload.clone(name=new_name)

        # Verify API call
        self.client.api.create_workload.assert_called_once()

        # Check that both arguments are passed as keyword arguments
        args, kwargs = cast(
            Tuple[Tuple[Any, ...], Dict[str, Any]],
            self.client.api.create_workload.call_args,
        )
        self.assertIn("config", kwargs)
        self.assertIn("metadata", kwargs)

        # Check metadata
        metadata = kwargs["metadata"]
        self.assertEqual(metadata["name"], new_name)
        # Original spec should be preserved
        self.assertEqual(metadata["spec"]["type"], self.attrs["spec"]["type"])
        self.assertEqual(
            metadata["spec"]["containers"], self.attrs["spec"]["containers"]
        )

    def test_clone_with_gvc(self) -> None:
        """Test clone method with custom gvc"""
        # Setup test data
        new_name: str = "cloned-workload"
        new_gvc: str = "new-gvc"

        # Mock the export method
        self.workload.export = MagicMock(
            return_value={
                "name": self.attrs["name"],
                "gvc": self.state["gvc"],
                "spec": self.attrs["spec"],
            }
        )

        # Mock successful API response
        mock_response: Mock = Mock(spec=Response)
        mock_response.status_code = 201
        mock_response.text = "Created"
        self.client.api.create_workload.return_value = mock_response

        # Mock print to avoid output during test
        with patch("builtins.print"):
            # Call clone method
            self.workload.clone(name=new_name, gvc=new_gvc)

        # Verify API call
        args, kwargs = cast(
            Tuple[Tuple[Any, ...], Dict[str, Any]],
            self.client.api.create_workload.call_args,
        )

        # Check that both arguments are passed as keyword arguments
        self.assertIn("config", kwargs)
        self.assertIn("metadata", kwargs)

        # Check config has new gvc
        self.assertEqual(kwargs["config"].gvc, new_gvc)

        # Check metadata has new gvc
        metadata = kwargs["metadata"]
        self.assertEqual(metadata["gvc"], new_gvc)

    def test_clone_with_type_change(self) -> None:
        """Test clone method with workload type change"""
        # Setup test data
        new_name: str = "cloned-workload"
        new_type: str = "serverless"  # Change from standard to serverless

        # Mock the export method
        self.workload.export = MagicMock(
            return_value={
                "name": self.attrs["name"],
                "gvc": self.state["gvc"],
                "spec": self.attrs["spec"],
            }
        )

        # Mock successful API response
        mock_response: Mock = Mock(spec=Response)
        mock_response.status_code = 201
        mock_response.text = "Created"
        self.client.api.create_workload.return_value = mock_response

        # Mock print to avoid output during test
        with patch("builtins.print"):
            # Call clone method
            self.workload.clone(name=new_name, workload_type=new_type)

        # Verify API call
        _, kwargs = cast(
            Tuple[Tuple[Any, ...], Dict[str, Any]],
            self.client.api.create_workload.call_args,
        )

        # Check metadata has new type
        metadata = kwargs["metadata"]
        self.assertEqual(metadata["spec"]["type"], new_type)

        # Check that autoscaling and capacityAI were updated for serverless type
        self.assertEqual(
            metadata["spec"]["defaultOptions"]["autoscaling"]["metric"], "cpu"
        )
        self.assertFalse(metadata["spec"]["defaultOptions"]["capacityAI"])

    def test_clone_error(self) -> None:
        """Test clone method with API error"""
        # Setup test data
        new_name: str = "cloned-workload"

        # Mock the export method
        self.workload.export = MagicMock(
            return_value={
                "name": self.attrs["name"],
                "gvc": self.state["gvc"],
                "spec": self.attrs["spec"],
            }
        )

        # Mock error API response
        mock_response: Mock = Mock(spec=Response)
        mock_response.status_code = 400
        mock_response.json.return_value = {"error": "Bad request"}
        mock_response.text = "Bad request"
        self.client.api.create_workload.return_value = mock_response

        # Mock print to avoid output during test
        with (
            patch("builtins.print"),
            self.assertRaises((RuntimeError, ValueError)),
        ):
            self.workload.clone(name=new_name)

        # Verify API call was attempted
        self.client.api.create_workload.assert_called_once()


class TestWorkloadCollection(unittest.TestCase):
    """Tests for the WorkloadCollection class"""

    def setUp(self) -> None:
        """Set up the test"""
        self.client: MagicMock = MagicMock()
        self.collection = WorkloadCollection(client=self.client)

    def test_model_attribute(self) -> None:
        """Test the model attribute is set correctly"""
        self.assertEqual(self.collection.model, Workload)

    def test_get_workload(self) -> None:
        """Test get method to retrieve a specific workload"""
        # Setup test data
        config: WorkloadConfig = WorkloadConfig(
            gvc="test-gvc", workload_id="test-workload"
        )
        workload_data: Dict[str, Any] = {
            "name": "test-workload",
            "spec": {"type": "standard"},
        }

        # Mock API response
        self.client.api.get_workload.return_value = workload_data

        # Call get method
        result = self.collection.get(config)

        # Verify API call
        self.client.api.get_workload.assert_called_once_with(config=config)

        # Verify result
        self.assertIsInstance(result, Workload)
        self.assertEqual(result.attrs, workload_data)
        self.assertEqual(result.state["gvc"], config.gvc)

    def test_list_with_gvc(self) -> None:
        """Test list method with gvc parameter"""
        # Setup test data
        gvc: str = "test-gvc"
        workloads: List[Dict[str, Any]] = [
            {"name": "workload1", "spec": {"type": "standard"}},
            {"name": "workload2", "spec": {"type": "serverless"}},
        ]

        # Mock API responses
        self.client.api.get_workload.side_effect = [
            {"items": workloads},  # First call returns list
            workloads[0],  # Second call returns workload1
            workloads[1],  # Third call returns workload2
        ]

        # Call list method
        result = self.collection.list(gvc=gvc)

        # Verify result structure
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)

        # Verify result contents
        self.assertIsInstance(result[0], Workload)
        self.assertEqual(result[0].attrs, workloads[0])
        self.assertEqual(result[0].state["gvc"], gvc)

        self.assertIsInstance(result[1], Workload)
        self.assertEqual(result[1].attrs, workloads[1])
        self.assertEqual(result[1].state["gvc"], gvc)

        # Verify API calls
        self.assertEqual(self.client.api.get_workload.call_count, 3)

    def test_list_with_config(self) -> None:
        """Test list method with config parameter"""
        # Setup test data
        config: WorkloadConfig = WorkloadConfig(gvc="test-gvc")
        workloads: List[Dict[str, Any]] = [
            {"name": "workload1", "spec": {"type": "standard"}},
            {"name": "workload2", "spec": {"type": "serverless"}},
        ]

        # Mock API responses
        self.client.api.get_workload.side_effect = [
            {"items": workloads},  # First call returns list
            workloads[0],  # Second call returns workload1
            workloads[1],  # Third call returns workload2
        ]

        # Call list method
        result = self.collection.list(config=config)

        # Verify result structure
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)

        # Verify API calls
        self.assertEqual(self.client.api.get_workload.call_count, 3)

        # Check first call was with the original config
        first_call = self.client.api.get_workload.call_args_list[0]

        # The config might be passed as a positional or keyword argument
        # Try keyword first
        if "config" in first_call[1]:
            self.assertEqual(first_call[1]["config"], config)
        else:
            # If not a keyword, check if it was passed as a positional argument
            self.assertEqual(first_call[0][0], config)

    def test_list_no_args(self) -> None:
        """Test list method with no arguments"""
        with self.assertRaises(ValueError):
            self.collection.list()

    @patch("cpln.models.workloads.get_default_workload_template")
    def test_create_standard(self, mock_template: Mock) -> None:
        """Test create method with standard workload type"""
        # Setup test data
        name: str = "test-standard-workload"
        gvc: str = "test-gvc"
        description: str = "Standard workload description"
        image: str = "nginx:latest"
        container_name: str = "app"
        workload_type: str = "standard"

        # Mock template
        template: Dict[str, Any] = {
            "name": "",
            "description": "",
            "spec": {
                "containers": [{"name": "", "image": ""}],
                "type": "standard",
                "defaultOptions": {
                    "autoscaling": {"metric": "memory"},
                    "capacityAI": True,
                },
            },
        }
        mock_template.return_value = template

        # Mock successful API response
        mock_response: Mock = Mock(spec=Response)
        mock_response.status_code = 201
        mock_response.text = "Created"
        self.client.api.create_workload.return_value = mock_response

        # Mock print to avoid output during test
        with patch("builtins.print"):
            # Call create method
            self.collection.create(
                name=name,
                gvc=gvc,
                description=description,
                image=image,
                container_name=container_name,
                workload_type=workload_type,
            )

        # Verify template was requested
        mock_template.assert_called_once_with(workload_type)

        # Verify API call
        self.client.api.create_workload.assert_called_once()

        # Check that both arguments are being passed
        # They may be passed as positional or keyword arguments
        args, kwargs = cast(
            Tuple[Tuple[Any, ...], Dict[str, Any]],
            self.client.api.create_workload.call_args,
        )

        if args and len(args) >= 1:
            # If passed as positional args
            self.assertEqual(args[0].gvc, gvc)
            if len(args) >= 2:
                metadata = args[1]
            else:
                self.assertIn("metadata", kwargs)
                metadata = kwargs["metadata"]
        else:
            # If passed as keyword args
            self.assertIn("config", kwargs)
            self.assertEqual(kwargs["config"].gvc, gvc)
            self.assertIn("metadata", kwargs)
            metadata = kwargs["metadata"]

        # Verify metadata content
        self.assertEqual(metadata["name"], name)
        self.assertEqual(metadata["description"], description)
        self.assertEqual(metadata["spec"]["containers"][0]["image"], image)
        self.assertEqual(metadata["spec"]["containers"][0]["name"], container_name)

    def test_create_missing_params(self) -> None:
        """Test create method with missing required parameters"""
        # Missing image
        with self.assertRaises(ValueError) as cm:
            self.collection.create(
                name="test-workload", gvc="test-gvc", container_name="app"
            )
        self.assertEqual(str(cm.exception), "Image is required.")

        # Missing container_name
        with self.assertRaises(ValueError) as cm:
            self.collection.create(
                name="test-workload", gvc="test-gvc", image="nginx:latest"
            )
        self.assertEqual(str(cm.exception), "Container name is required.")

        # No gvc or config provided
        with self.assertRaises(ValueError) as cm:
            self.collection.create(
                name="test-workload", image="nginx:latest", container_name="app"
            )
        self.assertEqual(
            str(cm.exception), "Either GVC or WorkloadConfig must be defined."
        )


if __name__ == "__main__":
    unittest.main()



import os
from unittest.mock import Mock

import pytest
from cpln.api.client import APIClient
from cpln.errors import APIError, NotFound


def test_api_client_initialization(mock_config):
    client = APIClient(config=mock_config)
    assert client.config == mock_config
    assert client.config.base_url == os.getenv("CPLN_BASE_URL")
    assert client.config.org == os.getenv("CPLN_ORG")
    assert client.config.token == os.getenv("CPLN_TOKEN")


def test_api_client_headers(mock_config):
    client = APIClient(config=mock_config)
    headers = client._headers
    assert headers == {"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}


def test_api_client_get_gvc(mock_api_client):
    # Mock response is already set up in the fixture
    mock_api_client._mock_get_response.status_code = 200
    mock_api_client._mock_get_response.json.return_value = {"data": "test"}

    # Call the API method
    result = mock_api_client.get_gvc()

    # Verify the get method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc"
    mock_api_client._mock_get.assert_called_once_with(
        expected_url, headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}
    )

    # Verify the result
    assert result == {"data": "test"}


def test_api_client_get_gvc_not_found(mock_api_client):
    # Set up the mock response for 404
    mock_api_client._mock_get_response.status_code = 404
    mock_api_client._mock_get_response.text = "Not Found"

    # Test that NotFound exception is raised
    with pytest.raises(NotFound):
        mock_api_client.get_gvc()

    # Verify the get method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc"
    mock_api_client._mock_get.assert_called_once_with(
        expected_url, headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}
    )


def test_api_client_get_gvc_error(mock_api_client):
    # Set up the mock response for 500
    mock_api_client._mock_get_response.status_code = 500
    mock_api_client._mock_get_response.text = "Internal Server Error"

    # Test that APIError exception is raised
    with pytest.raises(APIError):
        mock_api_client.get_gvc()

    # Verify the get method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc"
    mock_api_client._mock_get.assert_called_once_with(
        expected_url, headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}
    )


def test_api_client_get_image(mock_api_client):
    # Set up the mock response
    mock_api_client._mock_get_response.status_code = 200
    mock_api_client._mock_get_response.json.return_value = {"data": "test"}

    # Call the API method
    result = mock_api_client.get_image()

    # Verify the get method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/image"
    mock_api_client._mock_get.assert_called_once_with(
        expected_url, headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}
    )

    # Verify the result
    assert result == {"data": "test"}


def test_api_client_delete_gvc(mock_api_client):
    # Set up the mock response for successful delete
    mock_api_client._mock_delete_response.status_code = (
        204  # Success status code for DELETE
    )

    # Call the API method
    result = mock_api_client.delete_gvc("test-gvc")

    # Verify the delete method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc/test-gvc"
    mock_api_client._mock_delete.assert_called_once_with(
        expected_url, headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}
    )

    # Verify the result
    assert result == mock_api_client._mock_delete_response


def test_api_client_delete_gvc_not_found(mock_api_client):
    # Set up the mock response for 404
    mock_api_client._mock_delete_response.status_code = 404
    mock_api_client._mock_delete_response.text = "Not Found"

    # Test that NotFound exception is raised
    with pytest.raises(NotFound):
        mock_api_client.delete_gvc("nonexistent-gvc")

    # Verify the delete method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc/nonexistent-gvc"
    mock_api_client._mock_delete.assert_called_once_with(
        expected_url, headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"}
    )


def test_api_client_patch_workload(mock_api_client):
    # Set up the mock response for successful patch
    mock_api_client._mock_patch_response.status_code = 200

    # Test data
    test_data = {"key": "value"}

    # Create a mock workload config
    mock_config = Mock()
    mock_config.gvc = "test-gvc"
    mock_config.workload_id = "test-workload"

    # Call the API method
    result = mock_api_client.patch_workload(config=mock_config, data=test_data)

    # Verify the patch method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc/test-gvc/workload/test-workload"
    mock_api_client._mock_patch.assert_called_once_with(
        expected_url,
        json=test_data,
        headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"},
    )

    # Verify the result
    assert result == mock_api_client._mock_patch_response


def test_api_client_patch_workload_error(mock_api_client):
    # Set up the mock response for error
    mock_api_client._mock_patch_response.status_code = 400
    mock_api_client._mock_patch_response.text = "Bad Request"

    # Test data
    test_data = {"invalid": "data"}

    # Create a mock workload config
    mock_config = Mock()
    mock_config.gvc = "test-gvc"
    mock_config.workload_id = "test-workload"

    # Test that APIError exception is raised
    with pytest.raises(APIError):
        mock_api_client.patch_workload(config=mock_config, data=test_data)

    # Verify the patch method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/gvc/test-gvc/workload/test-workload"
    mock_api_client._mock_patch.assert_called_once_with(
        expected_url,
        json=test_data,
        headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"},
    )


def test_api_client_post(mock_api_client):
    # Set up the mock response for successful post
    mock_api_client._mock_post_response.status_code = (
        201  # Success status code for POST
    )

    # Test data
    test_data = {"name": "test-resource"}

    # Call the internal API method
    result = mock_api_client._post("test-endpoint", test_data)

    # Verify the post method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/test-endpoint"
    mock_api_client._mock_post.assert_called_once_with(
        expected_url,
        json=test_data,
        headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"},
    )

    # Verify the result
    assert result == mock_api_client._mock_post_response


def test_api_client_post_error(mock_api_client):
    # Set up the mock response for error
    mock_api_client._mock_post_response.status_code = 422
    mock_api_client._mock_post_response.text = "Unprocessable Entity"
    # Provide a json method that will be called in error handling
    mock_api_client._mock_post_response.json.return_value = {
        "error": "Validation failed"
    }

    # Test data
    test_data = {"invalid": "data"}

    # Test that APIError exception is raised
    with pytest.raises(APIError):
        mock_api_client._post("test-endpoint", test_data)

    # Verify the post method was called with the correct arguments
    expected_url = f"{os.getenv('CPLN_BASE_URL', 'https://api.cpln.io')}/org/{os.getenv('CPLN_ORG', 'ledgestone')}/test-endpoint"
    mock_api_client._mock_post.assert_called_once_with(
        expected_url,
        json=test_data,
        headers={"Authorization": f"Bearer {os.getenv('CPLN_TOKEN')}"},
    )



"""
Tests for utility functions.
"""

from examples.utils import safe_get_attr


class TestObject:
    """Test class with attributes in different case formats."""

    def __init__(self):
        self.camelCase = "camel"
        self.snake_case = "snake"
        self.TitleCase = "title"
        self.original = "original"


def test_safe_get_attr_original():
    """Test getting attribute in original format."""
    obj = TestObject()
    assert safe_get_attr(obj, "original") == "original"


def test_safe_get_attr_camel_case():
    """Test getting attribute in camelCase format."""
    obj = TestObject()
    assert safe_get_attr(obj, "camel_case") == "camel"
    assert safe_get_attr(obj, "camelCase") == "camel"


def test_safe_get_attr_snake_case():
    """Test getting attribute in snake_case format."""
    obj = TestObject()
    assert safe_get_attr(obj, "snakeCase") == "snake"
    assert safe_get_attr(obj, "snake_case") == "snake"


def test_safe_get_attr_title_case():
    """Test getting attribute in TitleCase format."""
    obj = TestObject()
    assert safe_get_attr(obj, "title_case") == "title"
    assert safe_get_attr(obj, "TitleCase") == "title"


def test_safe_get_attr_default():
    """Test getting non-existent attribute returns default value."""
    obj = TestObject()
    assert safe_get_attr(obj, "non_existent") == "N/A"
    assert safe_get_attr(obj, "nonExistent") == "N/A"
    assert safe_get_attr(obj, "NonExistent") == "N/A"


def test_safe_get_attr_custom_default():
    """Test getting non-existent attribute with custom default value."""
    obj = TestObject()
    assert safe_get_attr(obj, "non_existent", default="custom") == "custom"


def test_safe_get_attr_none_value():
    """Test getting attribute that exists but has None value."""
    obj = TestObject()
    obj.none_value = None
    assert safe_get_attr(obj, "none_value") == "None"
    assert safe_get_attr(obj, "noneValue") == "None"


def test_safe_get_attr_complex_object():
    """Test getting attribute with complex object value."""
    obj = TestObject()
    obj.complex = {"key": "value"}
    assert safe_get_attr(obj, "complex") == "{'key': 'value'}"
    assert safe_get_attr(obj, "Complex") == "{'key': 'value'}"


def test_safe_get_attr_empty_string():
    """Test getting attribute with empty string value."""
    obj = TestObject()
    obj.empty = ""
    assert safe_get_attr(obj, "empty") == ""
    assert safe_get_attr(obj, "Empty") == ""


def test_safe_get_attr_special_chars():
    """Test getting attribute with special characters in name."""
    obj = TestObject()
    obj.special_chars = "special"
    assert safe_get_attr(obj, "special-chars") == "special"
    assert safe_get_attr(obj, "specialChars") == "special"
    assert safe_get_attr(obj, "SpecialChars") == "special"



from typing import Optional

from ..config import WorkloadConfig
from ..errors import WebSocketExitCodeError
from ..utils import get_default_workload_template, load_template
from .resource import Collection, Model
from .workload_specs import WorkloadSpecState


class Workload(Model):
    """
    A workload on the server.
    """

    def get(self) -> dict[str, any]:
        """
        Get the workload.

        Returns:
            (dict): The workload.

        Raises:
            :py:class:`cpln.errors.NotFound`
                If the workload does not exist.
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.get_workload(self.config())

    def get_spec_state(self) -> WorkloadSpecState:
        self.get()
        return WorkloadSpecState.parse_from_spec(self.spec)

    def delete(self) -> None:
        """
        Delete the workload.

        Raises:
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        print(f"Deleting Workload: {self}")
        self.client.api.delete_workload(self.config())
        print("Deleted!")

    def clone(
        self,
        name: str,
        gvc: Optional[str] = None,
        workload_type: Optional[str] = None,
    ) -> None:
        """
        Clone the workload.

        Args:
            name (str): The name of the new workload.
            gvc (str, optional): The GVC to create the workload in. Defaults to None.
            workload_type (str, optional): The type of workload to create. Defaults to None.

        Returns:
            None

        Raises:
            Exception: If the API returns a non-2xx status code.
        """
        metadata = self.export()

        # TODO: I need to get identity link from the REST API, in order
        # to change it in the metadata. The path to the identity link is
        # different for different GVCs.

        # TODO: The parameters to the created/cloned workloads are too limited.
        # In order for this package to be more widely used, we need to implement
        # a way to pass more workload configuration parameters.
        metadata["name"] = name
        if gvc is not None:
            metadata["gvc"] = gvc

        if workload_type is not None:
            metadata["spec"]["type"] = workload_type

            # Ensure defaultOptions exists
            if "defaultOptions" not in metadata["spec"]:
                metadata["spec"]["defaultOptions"] = {}

            # Ensure autoscaling exists
            if "autoscaling" not in metadata["spec"]["defaultOptions"]:
                metadata["spec"]["defaultOptions"]["autoscaling"] = {}

            # Set autoscaling metric and capacityAI
            metadata["spec"]["defaultOptions"]["autoscaling"]["metric"] = "cpu"
            metadata["spec"]["defaultOptions"]["capacityAI"] = False

        response = self.client.api.create_workload(
            config=self.config(gvc=gvc),
            metadata=metadata,
        )
        if response.status_code // 100 == 2:
            print(response.status_code, response.text)
        else:
            print(response.status_code, response.json())
            raise RuntimeError(f"API call failed with status {response.status_code}")

    def suspend(self) -> None:
        self._change_suspend_state(state=True)

    def unsuspend(self) -> None:
        self._change_suspend_state(state=False)

    def exec(self, command: str, location: str):
        """
        Execute a command on the workload.

        Args:
            command (str): The command to execute.

        Returns:
            (dict): The response from the server.

        Raises:
            :py:class:`cpln.errors.WebSocketExitCodeError`
                If the command returns a non-zero exit code.
        """
        try:
            return self.client.api.exec_workload(
                config=self.config(location=location), command=command
            )
        except WebSocketExitCodeError as e:
            print(f"Command failed with exit code: {e}")
            raise

    def ping(self, location: Optional[str] = None) -> dict[str, any]:
        """
        Ping the workload.

        Args:
            location (str): The location of the workload.
                Default: None
        Returns:
            (dict): The response from the server containing status, message, and exit code.
        """
        try:
            self.exec(
                ["echo", "ping"],
                location=location,
            )
            return {
                "status": 200,
                "message": "Successfully pinged workload",
                "exit_code": 0,
            }
        except WebSocketExitCodeError as e:
            return {
                "status": 500,
                "message": f"Command failed with exit code: {e}",
                "exit_code": e.exit_code,
            }
        except Exception as e:
            return {"status": 500, "message": str(e), "exit_code": -1}

    def export(self) -> dict[str, any]:
        """
        Export the workload.
        """
        self.get()
        return {
            "name": self.attrs["name"],
            "gvc": self.state["gvc"],
            "spec": self.attrs["spec"],
        }

    def config(
        self, location: Optional[str] = None, gvc: Optional[str] = None
    ) -> WorkloadConfig:
        """
        Get the workload config.

        Args:
            location (str): The location of the workload.
                Default: None

        Returns:
            (WorkloadConfig): The workload config.
        """
        return WorkloadConfig(
            gvc=self.state["gvc"] if gvc is None else gvc,
            workload_id=self.attrs["name"],
            location=location,
        )

    def get_remote(self, location: Optional[str] = None) -> str:
        """
        Get the remote URL of the workload.

        Args:
            location (str): The location of the workload.
                Default: None

        Returns:
            (str): The remote of the workload.
        """
        return self.client.api.get_remote(self.config(location=location))

    def get_remote_wss(self, location: Optional[str] = None) -> str:
        """
        Get the remote WSS URL of the workload.

        Args:
            location (str): The location of the workload.
                Default: None

        Returns:
            (str): The remote WSS of the workload.
        """
        return self.client.api.get_remote_wss(self.config(location=location))

    def get_replicas(self, location: Optional[str] = None) -> list[str]:
        """
        Get the replicas of the workload.

        Args:
            location (str): The location of the workload.
                Default: None

        Returns:
            (list): The replicas of the workload.
        """
        return self.client.api.get_replicas(self.config(location=location))

    def get_containers(self, location: Optional[str] = None) -> list[str]:
        """
        Get the containers of the workload.

        Args:
            location (str): The location of the workload.
                Default: None

        Returns:
            (list): The containers of the workload.
        """
        return self.client.api.get_containers(self.config(location=location))

    def _change_suspend_state(self, state: bool = True) -> None:
        output = self.client.api.patch_workload(
            config=self.config(),
            data={"spec": {"defaultOptions": {"suspend": str(state).lower()}}},
        )
        print(f"{'' if state else 'Un'}Suspending Workload: {self}")
        return output


class WorkloadCollection(Collection):
    """
    Workloads on the server.
    """

    model = Workload

    def create(
        self,
        name: str,
        gvc: Optional[str] = None,
        config: Optional[WorkloadConfig] = None,
        description: Optional[str] = None,
        image: Optional[str] = None,
        container_name: Optional[str] = None,
        workload_type: Optional[str] = None,
        metadata_file_path: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> None:
        """
        Create the workload.
        """
        if gvc is None and config is None:
            raise ValueError("Either GVC or WorkloadConfig must be defined.")
        config = WorkloadConfig(gvc=gvc) if gvc else config

        if metadata is None:
            if metadata_file_path is None:
                if not image:
                    raise ValueError("Image is required.")
                if not container_name:
                    raise ValueError("Container name is required.")

                metadata = get_default_workload_template(
                    "serverless" if workload_type is None else workload_type
                )
                metadata["name"] = name
                metadata["description"] = description if description is not None else ""
                metadata["spec"]["containers"][0]["image"] = image
                metadata["spec"]["containers"][0]["name"] = container_name

            else:
                metadata = load_template(metadata_file_path)
        else:
            metadata["name"] = name
            if workload_type is not None:
                metadata["spec"]["type"] = workload_type
                metadata["spec"]["defaultOptions"]["autoscaling"]["metric"] = "cpu"
                metadata["spec"]["defaultOptions"]["capacityAI"] = False

        response = self.client.api.create_workload(config, metadata)
        if response.status_code // 100 == 2:
            print(response.status_code, response.text)
        else:
            print(response.status_code, response.json())
            raise RuntimeError(f"API call failed with status {response.status_code}")

    def get(self, config: WorkloadConfig):
        """
        Gets a workload.

        Args:
            config (WorkloadConfig): The workload config.

        Returns:
            (:py:class:`Workload`): The workload.

        Raises:
            :py:class:`cpln.errors.NotFound`
                If the workload does not exist.
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        return self.prepare_model(
            self.client.api.get_workload(config=config), state={"gvc": config.gvc}
        )

    def list(
        self, gvc: Optional[str] = None, config: Optional[WorkloadConfig] = None
    ) -> list[Workload]:
        """
        List workloads.

        Args:
            gvc (str): The GVC to list workloads from.
            config (WorkloadConfig): The workload config.

        Returns:
            (list): The workloads.

        Raises:
            ValueError: If neither gvc nor config is defined.
        """
        if gvc is None and config is None:
            raise ValueError("Either GVC or WorkloadConfig must be defined.")

        config = WorkloadConfig(gvc=gvc) if gvc else config
        resp = self.client.api.get_workload(config)["items"]
        return [
            self.get(
                config=WorkloadConfig(gvc=config.gvc, workload_id=workload["name"])
            )
            for workload in resp
        ]



from enum import Enum
from typing import Dict, Optional


class GenericExitCode(Enum):
    """Enumeration of generic exit codes and their meanings."""

    SUCCESS = 0
    GENERAL_ERROR = 1
    INVALID_ARGUMENTS = 2
    PERMISSION_DENIED = 3
    FILE_NOT_FOUND = 4
    PROCESS_INTERRUPTED = 130
    UNKNOWN_ERROR = 255

    @classmethod
    def get_message(cls, code: int) -> str:
        """Get the human-readable message for an exit code."""
        messages: Dict[int, str] = {
            cls.SUCCESS.value: "The command completed successfully.",
            cls.GENERAL_ERROR.value: "An error occurred while executing the command.",
            cls.INVALID_ARGUMENTS.value: "Invalid arguments were provided to the command.",
            cls.PERMISSION_DENIED.value: "Permission denied while executing the command.",
            cls.FILE_NOT_FOUND.value: "A required file or directory was not found.",
            cls.PROCESS_INTERRUPTED.value: "The process was interrupted by a signal.",
            cls.UNKNOWN_ERROR.value: "An unknown error occurred while executing the command.",
        }
        return messages.get(code, f"Unknown exit code: {code}")

    @classmethod
    def is_error(cls, code: int) -> bool:
        """Check if an exit code indicates an error."""
        return code != cls.SUCCESS.value

    @classmethod
    def get_error_type(cls, code: int) -> Optional[str]:
        """Get the type of error for an exit code."""
        error_types: Dict[int, str] = {
            cls.GENERAL_ERROR.value: "General Error",
            cls.INVALID_ARGUMENTS.value: "Invalid Arguments",
            cls.PERMISSION_DENIED.value: "Permission Denied",
            cls.FILE_NOT_FOUND.value: "File Not Found",
            cls.PROCESS_INTERRUPTED.value: "Process Interrupted",
            cls.UNKNOWN_ERROR.value: "Unknown Error",
        }
        return error_types.get(code)



import os
from pathlib import Path
from typing import Dict, Generator
from unittest.mock import MagicMock, Mock, patch

import pytest
import requests
from cpln import CPLNClient
from cpln.api.client import APIClient
from cpln.api.config import APIConfig
from dotenv import load_dotenv

# Load environment variables from .env file
env_path = Path(__file__).parent.parent.parent.parent / ".env"
load_dotenv(env_path)


@pytest.fixture(autouse=True)
def mock_env_vars() -> Generator[None, None, None]:
    # Only override if not set in .env
    env_vars: Dict[str, str] = {
        "CPLN_TOKEN": os.getenv("CPLN_TOKEN", "test-token"),
        "CPLN_ORG": os.getenv("CPLN_ORG", "test-org"),
        "CPLN_BASE_URL": os.getenv("CPLN_BASE_URL", "https://api.cpln.io"),
    }
    with patch.dict(os.environ, env_vars):
        yield


@pytest.fixture
def mock_config() -> APIConfig:
    return APIConfig(
        base_url=os.getenv("CPLN_BASE_URL"),
        org=os.getenv("CPLN_ORG"),
        token=os.getenv("CPLN_TOKEN"),
    )


@pytest.fixture
def mock_session() -> MagicMock:
    """
    Create a mock session with appropriate response behaviors.
    """
    mock_session: MagicMock = MagicMock(spec=requests.Session)

    # Set up default responses for HTTP methods
    get_response: Mock = Mock()
    get_response.status_code = 200
    get_response.json.return_value = {}
    mock_session.get.return_value = get_response

    post_response: Mock = Mock()
    post_response.status_code = 201
    post_response.json.return_value = {}
    post_response.text = "Created"
    mock_session.post.return_value = post_response

    patch_response: Mock = Mock()
    patch_response.status_code = 200
    patch_response.json.return_value = {}
    patch_response.text = "OK"
    mock_session.patch.return_value = patch_response

    delete_response: Mock = Mock()
    delete_response.status_code = 204
    delete_response.text = ""
    mock_session.delete.return_value = delete_response

    return mock_session


@pytest.fixture
def mock_api_client(
    mock_config: APIConfig, mock_session: MagicMock
) -> Generator[APIClient, None, None]:
    """
    Create a properly mocked APIClient instance.

    This fixture creates an APIClient with a patched requests.Session constructor
    to avoid initialization issues.
    """
    # Create a patcher for the requests.Session constructor
    with patch("requests.Session", return_value=mock_session):
        # Create an instance of APIClient with the mock config
        client: APIClient = APIClient.__new__(APIClient, **mock_config.asdict())

        # Make sure our session was properly set
        client.session = mock_session

        yield client


@pytest.fixture
def mock_cpln_client(mock_api_client: APIClient) -> Generator[CPLNClient, None, None]:
    """
    Create a properly mocked CPLNClient instance.

    This fixture patches the APIClient constructor to return our mock_api_client
    when CPLNClient creates it.
    """
    # Create a patch for the APIClient constructor
    with (
        patch("cpln.client.APIClient", return_value=mock_api_client),
        patch.object(CPLNClient, "__init__", return_value=None),
    ):
        # Create a CPLNClient instance
        client: CPLNClient = CPLNClient.__new__(CPLNClient)

        # Manually set the api attribute
        client.api = mock_api_client

        yield client



import unittest
from unittest.mock import MagicMock

from cpln.models.resource import Collection, Model


class TestModel(unittest.TestCase):
    def setUp(self):
        self.attrs = {"id": "test-id-123456789012", "name": "Test Model"}
        self.client = MagicMock()
        self.collection = MagicMock()
        self.model = Model(
            attrs=self.attrs, client=self.client, collection=self.collection
        )

    def test_init(self):
        """Test model initialization"""
        self.assertEqual(self.model.attrs, self.attrs)
        self.assertEqual(self.model.client, self.client)
        self.assertEqual(self.model.collection, self.collection)
        self.assertEqual(self.model.state, {})

    def test_init_defaults(self):
        """Test model initialization with defaults"""
        model = Model()
        self.assertEqual(model.attrs, {})
        self.assertIsNone(model.client)
        self.assertIsNone(model.collection)
        self.assertEqual(model.state, {})

    def test_id(self):
        """Test id property"""
        self.assertEqual(self.model.id, "test-id-123456789012")

    def test_short_id(self):
        """Test short_id property"""
        self.assertEqual(self.model.short_id, "test-id-1234")

    def test_label(self):
        """Test label property"""
        self.assertEqual(self.model.label, "Test Model")

    def test_repr(self):
        """Test string representation"""
        expected = f"<Model: {self.model.short_id} - {self.model.label}>"
        self.assertEqual(repr(self.model), expected)

    def test_eq(self):
        """Test equality comparison"""
        other = Model(attrs={"id": "test-id-123456789012"})
        self.assertEqual(self.model, other)

    def test_hash(self):
        """Test hashing"""
        expected_hash = hash(f"Model:{self.model.id}")
        self.assertEqual(hash(self.model), expected_hash)

    def test_reload(self):
        """Test reload method"""
        new_attrs = {"id": "test-id-123456789012", "name": "Updated Model"}
        self.collection.get.return_value = Model(attrs=new_attrs)
        self.model.reload()
        self.assertEqual(self.model.attrs, new_attrs)

    def test_attribute_access(self):
        """Test attribute access through __getattr__"""
        # Test accessing existing attribute
        self.assertEqual(self.model.name, "Test Model")

        # Test accessing non-existent attribute
        with self.assertRaises(AttributeError) as context:
            _ = self.model.non_existent
        self.assertEqual(
            str(context.exception), "'Model' has no attribute 'non_existent'"
        )

    def test_attribute_setting(self):
        """Test attribute setting through __setattr__"""
        # Test setting new attribute
        self.model.new_attr = "new value"
        self.assertEqual(self.model.attrs["new_attr"], "new value")

        # Test updating existing attribute
        self.model.name = "Updated Name"
        self.assertEqual(self.model.attrs["name"], "Updated Name")

        # Test setting protected attribute (should still work as it's not in protected_attrs)
        self.model.id = "new-id"
        self.assertEqual(self.model.attrs["id"], "new-id")

    def test_attribute_deletion(self):
        """Test attribute deletion through __delattr__"""
        # Test deleting existing attribute
        del self.model.name
        self.assertNotIn("name", self.model.attrs)

        # Test deleting non-existent attribute
        with self.assertRaises(AttributeError) as context:
            del self.model.non_existent
        self.assertEqual(
            str(context.exception), "'Model' has no attribute 'non_existent'"
        )

    def test_protected_attributes(self):
        """Test protected attributes behavior"""
        # Test that protected attributes can be set
        self.model.attrs = {"new": "value"}
        self.assertEqual(self.model.attrs, {"new": "value"})

        # Test that protected attributes can be accessed
        self.assertEqual(self.model.attrs, {"new": "value"})

        # Test that protected attributes cannot be deleted
        with self.assertRaises(AttributeError) as context:
            del self.model.attrs
        self.assertEqual(
            str(context.exception), "Cannot delete protected attribute 'attrs'"
        )


class TestCollection(unittest.TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.collection = Collection(client=self.client)

    def test_init(self):
        """Test collection initialization"""
        self.assertEqual(self.collection.client, self.client)

    def test_call(self):
        """Test __call__ method raises TypeError"""
        with self.assertRaises(TypeError):
            self.collection()

    def test_list_not_implemented(self):
        """Test list method raises NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self.collection.list()

    def test_get_not_implemented(self):
        """Test get method raises NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self.collection.get("key")

    def test_create_not_implemented(self):
        """Test create method raises NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self.collection.create()

    def test_prepare_model_with_model(self):
        """Test prepare_model with Model instance"""
        model = Model()
        prepared = self.collection.prepare_model(model)
        self.assertEqual(prepared, model)
        self.assertEqual(prepared.client, self.client)
        self.assertEqual(prepared.collection, self.collection)

    def test_prepare_model_with_dict(self):
        """Test prepare_model with dictionary"""

        class TestModel(Model):
            pass

        class TestCollection(Collection):
            model = TestModel

        collection = TestCollection(client=self.client)
        attrs = {"id": "test-id", "name": "Test Model"}
        prepared = collection.prepare_model(attrs)
        self.assertIsInstance(prepared, TestModel)
        self.assertEqual(prepared.attrs, attrs)
        self.assertEqual(prepared.client, self.client)
        self.assertEqual(prepared.collection, collection)

    def test_prepare_model_invalid_type(self):
        """Test prepare_model with invalid type"""
        with self.assertRaises(ValueError):
            self.collection.prepare_model("invalid")


if __name__ == "__main__":
    unittest.main()



import os
from unittest.mock import Mock

import pytest
import requests
from cpln.api.client import APIClient
from cpln.api.config import APIConfig


@pytest.fixture
def mock_config() -> Mock:
    """
    Return a mock APIConfig object for testing with properly configured properties.

    This provides a mock configuration object with default values for the base URL,
    organization name, and API token that can be used in tests.
    """
    mock = Mock(spec=APIConfig)
    mock.base_url = os.getenv("CPLN_BASE_URL", "https://api.cpln.io")
    mock.org = os.getenv("CPLN_ORG", "ledgestone")
    mock.token = (
        os.getenv("CPLN_TOKEN") or "mock-token"
    )  # default mock token if not set

    # Generate the org_url property
    mock.org_url = f"{mock.base_url}/org/{mock.org}"

    return mock


@pytest.fixture
def mock_api_client(mock_config: Mock) -> APIClient:
    """
    Create a properly mocked APIClient for testing.

    This fixture:
    1. Creates mock response objects for each HTTP method
    2. Sets up default successful responses
    3. Creates mock HTTP methods that return the mock responses
    4. Replaces the requests.Session HTTP methods with our mocks
    5. Stores both the mock methods and responses on the client for easy access in tests
    """
    # Create the mock response objects
    mock_get_response = Mock(spec=requests.Response)
    mock_post_response = Mock(spec=requests.Response)
    mock_patch_response = Mock(spec=requests.Response)
    mock_delete_response = Mock(spec=requests.Response)

    # Set up default successful responses
    mock_get_response.status_code = 200
    mock_get_response.json.return_value = {"data": "test"}

    mock_post_response.status_code = 201
    mock_post_response.json.return_value = {"data": "test"}

    mock_patch_response.status_code = 200
    mock_patch_response.json.return_value = {"data": "test"}

    mock_delete_response.status_code = 204

    # Create mock methods
    mock_get = Mock(return_value=mock_get_response)
    mock_post = Mock(return_value=mock_post_response)
    mock_patch = Mock(return_value=mock_patch_response)
    mock_delete = Mock(return_value=mock_delete_response)

    # Create the client with the mock config
    client = APIClient(config=mock_config)

    # Replace the requests.Session methods with our mocks
    client.get = mock_get  # type: ignore
    client.post = mock_post  # type: ignore
    client.patch = mock_patch  # type: ignore
    client.delete = mock_delete  # type: ignore

    # Store mocks on the client for access in tests
    client._mock_get = mock_get
    client._mock_post = mock_post
    client._mock_patch = mock_patch
    client._mock_delete = mock_delete

    # Store responses for direct access in tests
    client._mock_get_response = mock_get_response
    client._mock_post_response = mock_post_response
    client._mock_patch_response = mock_patch_response
    client._mock_delete_response = mock_delete_response

    return client



import os

import pytest
from cpln.api.config import APIConfig
from cpln.constants import (
    DEFAULT_CPLN_API_URL,
    DEFAULT_CPLN_API_VERSION,
    DEFAULT_TIMEOUT_SECONDS,
)


def test_api_config_default_values():
    config = APIConfig(
        base_url=os.getenv("CPLN_BASE_URL"),
        org=os.getenv("CPLN_ORG"),
        token=os.getenv("CPLN_TOKEN"),
    )
    assert config.base_url == DEFAULT_CPLN_API_URL
    assert config.org == os.getenv("CPLN_ORG")
    assert config.token == os.getenv("CPLN_TOKEN")
    assert config.version == DEFAULT_CPLN_API_VERSION
    assert config.timeout is DEFAULT_TIMEOUT_SECONDS


def test_api_config_custom_values():
    config = APIConfig(
        base_url=os.getenv("CPLN_BASE_URL"),
        org=os.getenv("CPLN_ORG"),
        token=os.getenv("CPLN_TOKEN"),
        version="2.0.0",
        timeout=30,
    )
    assert config.version == "2.0.0"
    assert config.timeout == 30


def test_api_config_org_url():
    config = APIConfig(
        base_url=os.getenv("CPLN_BASE_URL"),
        org=os.getenv("CPLN_ORG"),
        token=os.getenv("CPLN_TOKEN"),
    )
    expected_url = f"{os.getenv('CPLN_BASE_URL')}/org/{os.getenv('CPLN_ORG')}"
    assert config.org_url == expected_url


def test_api_config_missing_required_fields():
    with pytest.raises(TypeError):
        APIConfig()  # Missing required fields

    with pytest.raises(TypeError):
        APIConfig(base_url=os.getenv("CPLN_BASE_URL"))  # Missing org and token

    with pytest.raises(TypeError):
        APIConfig(
            base_url=os.getenv("CPLN_BASE_URL"), org=os.getenv("CPLN_ORG")
        )  # Missing token


def test_api_config_asdict():
    config = APIConfig(
        base_url=os.getenv("CPLN_BASE_URL"),
        org=os.getenv("CPLN_ORG"),
        token=os.getenv("CPLN_TOKEN"),
    )
    config_dict = {
        "base_url": os.getenv("CPLN_BASE_URL"),
        "org": os.getenv("CPLN_ORG"),
        "token": os.getenv("CPLN_TOKEN"),
        "version": DEFAULT_CPLN_API_VERSION,
        "timeout": DEFAULT_TIMEOUT_SECONDS,
        "org_url": f"{os.getenv('CPLN_BASE_URL')}/org/{os.getenv('CPLN_ORG')}",
    }
    assert config.asdict() == config_dict



class ImageApiMixin:
    """
    A mixin class that provides image-related API methods.
    """

    def get_image(self, image_id: str = None):
        """
        Retrieves image information from the Control Plane API.

        Args:
            image_id (str, optional): The ID of the specific image to retrieve.
                If not provided, returns a list of all images.

        Returns:
            dict: The image information or list of images

        Raises:
            APIError: If the request fails
        """
        endpoint = "image"
        if image_id:
            endpoint += f"/{image_id}"
        return self._get(endpoint)

    def delete_image(self, image_id):
        """
        Deletes an image from the Control Plane.

        Args:
            image_id (str): The ID of the image to delete

        Returns:
            requests.Response: The response from the API

        Raises:
            APIError: If the request fails
        """
        return self._delete(f"image/{image_id}")



from enum import Enum
from typing import Dict, Optional


class AwsExitCode(Enum):
    """Enumeration of common exit codes and their meanings."""

    SUCCESS = 0
    CONFIG_ERROR = 1
    PARSE_ERROR = 2
    SIGINT = 130
    INVALID_SYNTAX = 252
    INVALID_ENV = 253
    SERVICE_ERROR = 254
    GENERAL_ERROR = 255

    @classmethod
    def get_message(cls, code: int) -> str:
        """Get the human-readable message for an exit code."""
        messages: Dict[int, str] = {
            cls.SUCCESS.value: "The command was successful. There were no errors thrown by either the CLI or by the service that the request was made to.",
            cls.CONFIG_ERROR.value: "The configuration file parsed to the CLI was not found or might be corrupt.",
            cls.PARSE_ERROR.value: "The command entered on the command line failed to be parsed. Parsing failures can be caused by, but are not limited to, missing any required subcommands or arguments or using any unknown commands or arguments.",
            cls.SIGINT.value: "The process received a SIGINT (Ctrl-C).",
            cls.INVALID_SYNTAX.value: "Command syntax was invalid, an unknown parameter was provided, or a parameter value was incorrect and prevented the command from running.",
            cls.INVALID_ENV.value: "The system environment or configuration was invalid. While the command provided may be syntactically valid, missing configuration or credentials prevented the command from running.",
            cls.SERVICE_ERROR.value: "The command was successfully parsed and a request was made to the specified service but the service returned an error. This will generally indicate incorrect API usage or other service specific issues.",
            cls.GENERAL_ERROR.value: "General catch-all error. The command may have parsed correctly but an unspecified runtime error occurred when running the command.",
        }
        return messages.get(code, f"Unknown exit code: {code}")

    @classmethod
    def is_error(cls, code: int) -> bool:
        """Check if an exit code indicates an error."""
        return code != cls.SUCCESS.value

    @classmethod
    def get_error_type(cls, code: int) -> Optional[str]:
        """Get the type of error for an exit code."""
        error_types: Dict[int, str] = {
            cls.CONFIG_ERROR.value: "Configuration Error",
            cls.PARSE_ERROR.value: "Parse Error",
            cls.SIGINT.value: "Interrupted",
            cls.INVALID_SYNTAX.value: "Syntax Error",
            cls.INVALID_ENV.value: "Environment Error",
            cls.SERVICE_ERROR.value: "Service Error",
            cls.GENERAL_ERROR.value: "General Error",
        }
        return error_types.get(code)



class Model:
    """
    A base class for representing a single object on the server.
    """

    id_attribute = "id"
    label_attribute = "name"

    def __init__(self, attrs=None, client=None, collection=None, state=None):
        #: A client pointing at the server that this object is on.
        self.client = client

        #: The collection that this model is part of.
        self.collection = collection

        #: The state that represents this model.
        self.state = state
        if self.state is None:
            self.state = {}

        #: The raw representation of this object from the API
        self.attrs = attrs
        if self.attrs is None:
            self.attrs = {}

    def __repr__(self):
        short_id = self.short_id or "None"
        return f"<{self.__class__.__name__}: {short_id} - {self.label}>"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def __hash__(self):
        return hash(f"{self.__class__.__name__}:{self.id}")

    def __getattr__(self, name):
        """
        Allow attribute-style access to the attrs dictionary.

        Args:
            name: The name of the attribute to get

        Returns:
            The value of the attribute from attrs

        Raises:
            AttributeError: If the attribute doesn't exist in attrs
        """
        try:
            return self.attrs[name]
        except KeyError:
            raise AttributeError(
                f"'{self.__class__.__name__}' has no attribute '{name}'"
            ) from None

    def __setattr__(self, name, value):
        """
        Handle attribute assignment, storing values in attrs if they're not
        special attributes.

        Args:
            name: The name of the attribute to set
            value: The value to set the attribute to
        """
        # List of attributes that should be set directly on the instance
        direct_attrs = {"client", "collection", "state", "attrs"}

        if name in direct_attrs:
            super().__setattr__(name, value)
        else:
            # If attrs hasn't been initialized yet, initialize it
            if not hasattr(self, "attrs"):
                self.attrs = {}
            self.attrs[name] = value

    def __delattr__(self, name):
        """
        Handle attribute deletion, removing values from attrs if they're not
        special attributes.

        Args:
            name: The name of the attribute to delete

        Raises:
            AttributeError: If trying to delete a special attribute
        """
        # List of attributes that should not be deleted
        protected_attrs = {"client", "collection", "state", "attrs"}

        if name in protected_attrs:
            raise AttributeError(f"Cannot delete protected attribute '{name}'")

        if name in self.attrs:
            del self.attrs[name]
        else:
            raise AttributeError(
                f"'{self.__class__.__name__}' has no attribute '{name}'"
            )

    @property
    def id(self):
        """
        The ID of the object.
        """
        return self.attrs.get(self.id_attribute)

    @property
    def short_id(self):
        """
        The ID of the object, truncated to 12 characters.
        """
        if self.id is None:
            return None
        return self.id[:12]

    @property
    def label(self):
        """
        The label of the object.
        """
        return self.attrs.get(self.label_attribute)

    def reload(self):
        """
        Load this object from the server again and update ``attrs`` with the
        new data.
        """
        new_model = self.collection.get(self.id)
        self.attrs = new_model.attrs


class Collection:
    #: The type of object this collection represents, set by subclasses
    model = None

    def __init__(self, client=None):
        #: The client pointing at the server that this collection of objects
        #: is on.
        self.client = client

    def __call__(self, *args, **kwargs):
        raise TypeError(
            f"'{self.__class__.__name__}' object is not callable. "
            "You might be trying to use the old (pre-2.0) API - "
            "use docker.APIClient if so."
        )

    def list(self):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError

    def create(self, attrs=None):
        raise NotImplementedError

    def prepare_model(self, attrs, state=None):
        """
        Create a model from a set of attributes.
        """
        if isinstance(attrs, Model):
            attrs.client = self.client
            attrs.collection = self
            attrs.state = state
            return attrs
        elif isinstance(attrs, dict):
            return self.model(
                attrs=attrs, client=self.client, collection=self, state=state
            )
        else:
            model_name = self.model.__name__ if self.model else "Model"
            raise ValueError(f"Can't create {model_name} from {attrs}")



# import os
# from typing import cast

# from cpln import CPLNClient


# def test_cpln_client_initialization(mock_cpln_client: CPLNClient) -> None:
#     base_url = cast(str, os.getenv("CPLN_BASE_URL", "https://api.cpln.io"))
#     org = cast(str, os.getenv("CPLN_ORG", "test-org"))
#     token = cast(str, os.getenv("CPLN_TOKEN", "mock-token"))

#     assert mock_cpln_client.api.config.base_url == base_url
#     assert mock_cpln_client.api.config.org == org
#     assert mock_cpln_client.api.config.token == token


# def test_cpln_client_from_env() -> None:
#     with patch("cpln.utils.kwargs_from_env") as mock_kwargs:
#         base_url = cast(str, os.getenv("CPLN_BASE_URL", "https://api.cpln.io"))
#         org = cast(str, os.getenv("CPLN_ORG", "test-org"))
#         token = cast(str, os.getenv("CPLN_TOKEN", "mock-token"))

#         mock_kwargs.return_value = {"base_url": base_url, "org": org, "token": token}
#         client = CPLNClient.from_env()
#         assert client.api.config.base_url == base_url
#         assert client.api.config.org == org
#         assert client.api.config.token == token


# def test_cpln_client_gvcs_property(mock_cpln_client: CPLNClient) -> None:
#     gvcs = mock_cpln_client.gvcs
#     assert gvcs.client == mock_cpln_client


# def test_cpln_client_images_property(mock_cpln_client: CPLNClient) -> None:
#     images = mock_cpln_client.images
#     assert images.client == mock_cpln_client


# def test_cpln_client_workloads_property(mock_cpln_client: CPLNClient) -> None:
#     workloads = mock_cpln_client.workloads
#     assert workloads.client == mock_cpln_client



import json

from websocket import WebSocketApp

from ..errors import (
    WebSocketConnectionError,
    WebSocketExitCodeError,
    WebSocketOperationError,
)
from .exit_codes import AwsExitCode, GenericExitCode, PostgresExitCode


class WebSocketAPI:
    def __init__(self, remote_wss: str):
        self.remote_wss = remote_wss
        self._request = None
        self._error = None

    def exec(self, **kwargs):
        self._request = kwargs
        self._error = None
        ws = self.websocket()
        ws.run_forever()

        if self._error:
            raise self._error

        return self._request

    def websocket(self):
        ws = WebSocketApp(
            self.remote_wss,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
            on_open=self._on_open,
        )
        return ws

    def _on_message(self, ws: WebSocketApp, message: str):
        try:
            decoded_message = message.decode()
            exit_code = 0

            # Check for non-zero exit code
            if "exit code" in decoded_message.lower():
                try:
                    exit_code = decoded_message.split("exit code")[-1].strip()
                    exit_code = int(exit_code)
                except ValueError:
                    pass

            if exit_code != 0:
                command = self._request.get("command", [])
                if isinstance(command, list):
                    command = command[0] if command else ""

                if command == "aws":
                    if AwsExitCode.is_error(exit_code):
                        error_type = AwsExitCode.get_error_type(exit_code)
                        error_message = AwsExitCode.get_message(exit_code)
                elif command in ["pg_dump", "pg_restore"]:
                    if PostgresExitCode.is_error(exit_code):
                        error_type = PostgresExitCode.get_error_type(exit_code)
                        error_message = PostgresExitCode.get_message(exit_code, command)
                else:
                    if GenericExitCode.is_error(exit_code):
                        error_type = GenericExitCode.get_error_type(exit_code)
                        error_message = GenericExitCode.get_message(exit_code)

                self._error = WebSocketExitCodeError(
                    f"{error_type if error_type else 'Error'} (exit code {exit_code}): {error_message}\n"
                    f"Full message: {decoded_message}"
                )

                return exit_code

            # Check for error messages
            if "error" in decoded_message.lower():
                self._error = WebSocketOperationError(
                    f"Error in message: {decoded_message}"
                )
                return

            # Check for failure messages
            if "failed" in decoded_message.lower():
                self._error = WebSocketOperationError(
                    f"Operation failed: {decoded_message}"
                )
                return

            print(decoded_message)
            return exit_code

        except Exception as e:
            ws.sock.close()
            self._error = WebSocketOperationError(f"Error processing message: {str(e)}")

    def _on_error(self, ws: WebSocketApp, error: str):
        self._error = WebSocketConnectionError(f"WebSocket error: {error}")

    def _on_close(self, ws: WebSocketApp, close_status_code: int, close_msg: str):
        if not (
            close_status_code == 1000 or close_status_code is None
        ):  # 1000 is normal closure
            self._error = WebSocketConnectionError(
                f"Connection closed unexpectedly with code {close_status_code}: {close_msg}"
            )
        print(f"Connection closed, exit code: {close_status_code}")

    def _on_open(self, ws: WebSocketApp):
        print("Connection opened")
        try:
            ws.send(json.dumps(self._request, indent=4))
        except Exception as e:
            self._error = WebSocketConnectionError(
                f"Error sending initial request: {str(e)}"
            )
            ws.sock.close()



from typing import Optional

from .api import APIClient
from .models import GVCCollection, ImageCollection, WorkloadCollection
from .utils import kwargs_from_env


class CPLNClient:
    """
    A client for communicating with a Control Plane Server.

    Example:
        ```
        >>> import cpln
        >>> client = cpln.CPLNClient(base_url='https://api.cpln.io')
        ```

    Args:
        base_url (str): URL to the Control Plane server.
        org (str): The orgnanization namespace of your control plane service.
        token (str): Authorization token for accessing the use of the API.
    """

    def __init__(
        self,
        base_url: Optional[str] = None,
        org: Optional[str] = None,
        token: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize a CPLNClient.

        Args:
            base_url (str, optional): URL to the Control Plane server.
            org (str, optional): The organization namespace of your control plane service.
            token (str, optional): Authorization token for accessing the API.
            **kwargs: Additional arguments to pass to the APIClient.
        """
        # Create a config dict with all the args
        config_args = {}
        if base_url is not None:
            config_args["base_url"] = base_url
        if org is not None:
            config_args["org"] = org
        if token is not None:
            config_args["token"] = token

        # Add any other kwargs
        config_args.update(kwargs)

        # Create the API client
        self.api = APIClient(**config_args)

    @classmethod
    def from_env(cls, **kwargs):
        """
        Return a client configured from environment variables.

        The environment variables used are the same as those used by the
        cpln command-line client. They are:

        .. envvar:: `CPLN_TOKEN`
        Authorization token for accessing the use of the API.

        .. envvar:: `CPLN_ORG`
        The orgnanization namespace of your control plane service.

        .. envvar:: `CPLN_BASE_URL`
        URL to the Control Plane server.

        Args:
            version (str): The version of the API to use. Set to ``auto`` to
                automatically detect the server's version. Default: ``auto``
            timeout (int): Default timeout for API calls, in seconds.
            max_pool_size (int): The maximum number of connections
                to save in the pool.
            environment (dict): The environment to read environment variables
                from. Default: the value of ``os.environ``

        Example:

            >>> import cpln
            >>> client = cpln.from_env()

        Returns:
            CPLNClient: A client configured from environment variables.
        """
        # Get configuration from environment variables
        env_config = kwargs_from_env(**kwargs)

        # Create a new client instance with the environment configuration
        return cls(**env_config)

    @property
    def gvcs(self):
        """
        A collection of GVCs in the Control Plane.
        """
        return GVCCollection(client=self)

    @property
    def images(self):
        """
        A collection of images in the Control Plane.
        """
        return ImageCollection(client=self)

    @property
    def workloads(self):
        """
        A collection of workloads in the Control Plane.
        """
        return WorkloadCollection(client=self)


from_env = CPLNClient.from_env



from dataclasses import asdict, dataclass

from ..constants import (
    DEFAULT_CPLN_API_URL,
    DEFAULT_CPLN_API_VERSION,
    DEFAULT_TIMEOUT_SECONDS,
)


@dataclass
class APIConfig:
    """
    Configuration class for the Control Plane API client.

    This class holds the configuration parameters needed to interact with the Control Plane API,
    including authentication details, organization information, and API settings.

    Args:
        token (str): Authorization token for accessing the API
        org (str): Organization name in Control Plane
        base_url (str, optional): Base URL for the Control Plane API. Defaults to DEFAULT_CPLN_API_URL
        version (str, optional): API version to use. Defaults to DEFAULT_CPLN_API_VERSION
        timeout (int, optional): Request timeout in seconds. Defaults to DEFAULT_TIMEOUT_SECONDS
        org_url (str, optional): Organization-specific API URL. Will be automatically set based on base_url and org
    """

    token: str
    org: str
    base_url: str = DEFAULT_CPLN_API_URL
    version: str = DEFAULT_CPLN_API_VERSION
    timeout: int = DEFAULT_TIMEOUT_SECONDS
    org_url: str = None

    def __post_init__(self):
        """
        Post-initialization hook that sets the organization URL.
        """
        self.org_url = self.get_org_url()

    def get_org_url(self) -> str:
        """
        Constructs and returns the organization-specific API URL.

        Returns:
            str: The complete URL for the organization's API endpoint
        """
        return f"{self.base_url}/org/{self.org}"

    def asdict(self):
        """
        Converts the configuration object to a dictionary.

        Returns:
            dict: A dictionary containing all configuration parameters
        """
        return asdict(self)



from typing import Any, Dict
from unittest.mock import MagicMock, Mock, patch

import pytest
from cpln.api.config import APIConfig
from cpln.api.workload import WorkloadApiMixin, WorkloadDeploymentMixin
from cpln.config import WorkloadConfig


class TestWorkloadDeploymentMixin:
    """Tests for the WorkloadDeploymentMixin class"""

    def setup_method(self) -> None:
        """Set up the test"""
        self.mixin: WorkloadDeploymentMixin = WorkloadDeploymentMixin()
        self.mixin._get = MagicMock()
        self.mixin.config = MagicMock(spec=APIConfig)
        self.mixin.config.token = "test-token"
        self.mixin.config.org = "test-org"
        self.mixin.config.asdict.return_value = {
            "base_url": "https://api.cpln.io",
            "token": "test-token",
            "org": "test-org",
        }

        self.config: WorkloadConfig = WorkloadConfig(
            gvc="test-gvc", workload_id="test-workload", location="test-location"
        )

    def test_get_workload_deployment(self) -> None:
        """Test get_workload_deployment method"""
        deployment_data: Dict[str, Any] = {
            "status": {
                "remote": "https://test-remote",
                "versions": [{"containers": {"container1": {}, "container2": {}}}],
            }
        }
        self.mixin._get.return_value = deployment_data
        result = self.mixin.get_workload_deployment(self.config)

        self.mixin._get.assert_called_once_with(
            "gvc/test-gvc/workload/test-workload/deployment/test-location"
        )
        assert result == deployment_data

    def test_get_workload_deployment_invalid_config(self) -> None:
        """Test get_workload_deployment with invalid config"""
        invalid_config: WorkloadConfig = WorkloadConfig(
            gvc="test-gvc", workload_id=None
        )

        with pytest.raises(ValueError, match="Config not set properly"):
            self.mixin.get_workload_deployment(invalid_config)

    @patch.object(WorkloadDeploymentMixin, "__new__")
    def test_get_remote_api(self, mock_new: patch) -> None:
        """Test get_remote_api method"""
        api_config: APIConfig = APIConfig(
            base_url="https://api.cpln.io", org="test-org", token="test-token"
        )

        # Mock instance returned by __new__
        mock_instance: MagicMock = MagicMock()
        mock_instance.get_remote.return_value = "https://test-remote"
        mock_new.return_value = mock_instance

        # Call get_remote_api
        result = WorkloadDeploymentMixin.get_remote_api(api_config, self.config)

        # Verify the instance was created with the right args
        mock_new.assert_called()
        # Verify get_remote was called
        mock_instance.get_remote.assert_called_once_with(self.config)
        # The result should be the mock_instance
        assert result is mock_instance

    def test_get_remote(self) -> None:
        """Test get_remote method"""
        deployment_data: Dict[str, Any] = {"status": {"remote": "https://test-remote"}}
        self.mixin.get_workload_deployment = MagicMock(return_value=deployment_data)

        result = self.mixin.get_remote(self.config)

        self.mixin.get_workload_deployment.assert_called_once_with(self.config)
        assert result == "https://test-remote"

    def test_get_remote_wss(self) -> None:
        """Test get_remote_wss method"""
        self.mixin.get_remote = MagicMock(return_value="https://test-remote")

        result = self.mixin.get_remote_wss(self.config)

        self.mixin.get_remote.assert_called_once_with(self.config)
        assert result == "wss://test-remote/remote"

    def test_get_replicas(self) -> None:
        """Test get_replicas method"""
        # Mock the remote API client
        mock_remote_api: MagicMock = MagicMock()
        mock_remote_api._get.return_value = {"items": ["replica1", "replica2"]}
        self.mixin.get_remote_api = MagicMock(return_value=mock_remote_api)

        result = self.mixin.get_replicas(self.config)

        self.mixin.get_remote_api.assert_called_once_with(
            self.mixin.config, self.config
        )
        mock_remote_api._get.assert_called_once_with(
            "/gvc/test-gvc/workload/test-workload"
        )
        assert result == ["replica1", "replica2"]

    def test_get_containers(self) -> None:
        """Test get_containers method"""
        deployment_data: Dict[str, Any] = {
            "status": {
                "versions": [
                    {
                        "containers": {
                            "container1": {},
                            "container2": {},
                            "cpln-mounter": {},  # This should be ignored
                        }
                    }
                ]
            }
        }
        self.mixin.get_workload_deployment = MagicMock(return_value=deployment_data)

        result = self.mixin.get_containers(self.config)

        self.mixin.get_workload_deployment.assert_called_once_with(self.config)
        assert sorted(result) == ["container1", "container2"]
        assert "cpln-mounter" not in result


class TestWorkloadApiMixin:
    """Tests for the WorkloadApiMixin class"""

    def setup_method(self) -> None:
        """Set up the test"""
        self.mixin: WorkloadApiMixin = WorkloadApiMixin()
        self.mixin._get = MagicMock()
        self.mixin._post = MagicMock()
        self.mixin._delete = MagicMock()
        self.mixin._patch = MagicMock()

        # Since WorkloadApiMixin inherits from WorkloadDeploymentMixin,
        # we need to mock those methods too
        self.mixin.get_containers = MagicMock(return_value=["container1"])
        self.mixin.get_replicas = MagicMock(return_value=["replica1"])
        self.mixin.get_remote_wss = MagicMock(return_value="wss://test-remote")

        self.mixin.config = MagicMock(spec=APIConfig)
        self.mixin.config.token = "test-token"
        self.mixin.config.org = "test-org"

        self.config: WorkloadConfig = WorkloadConfig(
            gvc="test-gvc", workload_id="test-workload", location="test-location"
        )

    def test_get_workload_with_id(self) -> None:
        """Test get_workload method with workload ID"""
        self.mixin._get.return_value = {"name": "test-workload"}

        result = self.mixin.get_workload(self.config)

        self.mixin._get.assert_called_once_with("gvc/test-gvc/workload/test-workload")
        assert result == {"name": "test-workload"}

    def test_get_workload_without_id(self) -> None:
        """Test get_workload method without workload ID"""
        config: WorkloadConfig = WorkloadConfig(gvc="test-gvc")
        self.mixin._get.return_value = {
            "items": [{"name": "workload1"}, {"name": "workload2"}]
        }

        result = self.mixin.get_workload(config)

        self.mixin._get.assert_called_once_with("gvc/test-gvc/workload")
        assert result == {"items": [{"name": "workload1"}, {"name": "workload2"}]}

    def test_create_workload(self) -> None:
        """Test create_workload method"""
        metadata: Dict[str, str] = {
            "name": "new-workload",
            "description": "Test workload",
        }
        mock_response: Mock = Mock()
        self.mixin._post.return_value = mock_response

        result = self.mixin.create_workload(self.config, metadata)

        self.mixin._post.assert_called_once_with("gvc/test-gvc/workload", data=metadata)
        assert result == mock_response

    def test_delete_workload(self) -> None:
        """Test delete_workload method"""
        mock_response: Mock = Mock()
        self.mixin._delete.return_value = mock_response

        result = self.mixin.delete_workload(self.config)

        self.mixin._delete.assert_called_once_with(
            "gvc/test-gvc/workload/test-workload"
        )
        assert result == mock_response

    def test_patch_workload(self) -> None:
        """Test patch_workload method"""
        data: Dict[str, Any] = {"spec": {"defaultOptions": {"suspend": "true"}}}
        mock_response: Mock = Mock()
        self.mixin._patch.return_value = mock_response

        result = self.mixin.patch_workload(self.config, data)

        self.mixin._patch.assert_called_once_with(
            "gvc/test-gvc/workload/test-workload", data=data
        )
        assert result == mock_response

    @patch("cpln.api.workload.WebSocketAPI")
    def test_exec_workload(self, mock_websocket_api: patch) -> None:
        """Test exec_workload method"""
        command: str = "echo 'Hello, World!'"
        mock_instance: Mock = Mock()
        mock_websocket_api.return_value = mock_instance
        mock_instance.exec.return_value = {"output": "Hello, World!"}

        result = self.mixin.exec_workload(self.config, command)

        # Verify we get the containers, replicas, and remote URL
        self.mixin.get_containers.assert_called_once_with(self.config)
        self.mixin.get_replicas.assert_called_once_with(self.config)
        self.mixin.get_remote_wss.assert_called_once_with(self.config)

        # Verify we create the WebSocketAPI and call exec
        mock_websocket_api.assert_called_once_with("wss://test-remote")
        mock_instance.exec.assert_called_once_with(
            token="test-token",
            org="test-org",
            gvc="test-gvc",
            container="container1",
            pod="replica1",
            command=["echo", "'Hello,", "World!'"],
        )
        assert result == {"output": "Hello, World!"}



import unittest
from unittest.mock import MagicMock

from cpln.models.gvcs import GVC, GVCCollection


class TestGVC(unittest.TestCase):
    def setUp(self):
        self.attrs = {"name": "test-gvc", "description": "Test GVC Description"}
        self.client = MagicMock()
        self.collection = MagicMock()
        self.gvc = GVC(attrs=self.attrs, client=self.client, collection=self.collection)

    def test_get(self):
        """Test get method"""
        expected_response = {"name": "test-gvc"}
        self.client.api.get_gvc.return_value = expected_response
        result = self.gvc.get()
        self.assertEqual(result, expected_response)
        self.client.api.get_gvc.assert_called_once_with(self.attrs["name"])

    def test_create(self):
        """Test create method"""
        self.gvc.create()
        self.client.api.create_gvc.assert_called_once_with(
            self.attrs["name"], self.attrs["description"]
        )

    def test_delete(self):
        """Test delete method"""
        self.gvc.delete()
        self.client.api.delete_gvc.assert_called_once_with(self.attrs["name"])


class TestGVCCollection(unittest.TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.collection = GVCCollection(client=self.client)

    def test_get(self):
        """Test get method"""
        name = "test-gvc"
        expected_gvc = {"name": name}
        self.client.api.get_gvc.return_value = expected_gvc
        result = self.collection.get(name)
        self.assertIsInstance(result, GVC)
        self.assertEqual(result.attrs, expected_gvc)
        self.client.api.get_gvc.assert_called_once_with(name)

    def test_list(self):
        """Test list method"""
        response = {"items": [{"name": "gvc1"}, {"name": "gvc2"}]}
        self.client.api.get_gvc.return_value = response
        result = self.collection.list()
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], GVC)
        self.assertIsInstance(result[1], GVC)
        self.assertEqual(result[0].attrs["name"], "gvc1")
        self.assertEqual(result[1].attrs["name"], "gvc2")
        self.client.api.get_gvc.assert_called_once_with()


if __name__ == "__main__":
    unittest.main()



import json
import unittest
from typing import Dict, List
from unittest.mock import MagicMock, patch

from cpln.errors import (
    WebSocketConnectionError,
    WebSocketExitCodeError,
    WebSocketOperationError,
)
from cpln.utils.websocket import WebSocketAPI


class TestWebSocketAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.remote_wss: str = "wss://test.example.com"
        self.ws_api: WebSocketAPI = WebSocketAPI(self.remote_wss)
        self.mock_ws: MagicMock = MagicMock()

    # @patch('websocket.WebSocketApp')
    # def test_websocket_initialization(self, mock_ws_app):
    #     """Test that WebSocketApp is initialized correctly"""
    #     self.ws_api.websocket()
    #     mock_ws_app.assert_called_once_with(
    #         self.remote_wss,
    #         on_message=self.ws_api._on_message,
    #         on_error=self.ws_api._on_error,
    #         on_close=self.ws_api._on_close,
    #         on_open=self.ws_api._on_open,
    #     )

    def test_on_open_success(self) -> None:
        """Test successful connection opening"""
        request: Dict[str, str] = {"command": "test"}
        self.ws_api._request = request
        self.ws_api._on_open(self.mock_ws)
        self.mock_ws.send.assert_called_once_with(json.dumps(request, indent=4))

    def test_on_open_error(self) -> None:
        """Test error during connection opening"""
        self.ws_api._request: Dict[str, str] = {"command": "test"}
        self.mock_ws.send.side_effect = Exception("Connection error")
        self.ws_api._on_open(self.mock_ws)
        self.assertIsInstance(self.ws_api._error, WebSocketConnectionError)

    def test_on_message_success(self) -> None:
        """Test successful message processing"""
        message: bytes = b"Test message"
        self.ws_api._on_message(self.mock_ws, message)
        self.assertIsNone(self.ws_api._error)

    def test_on_message_exit_code_error(self) -> None:
        """Test handling of exit code errors"""
        message: bytes = b"exit code 1"
        self.ws_api._request: Dict[str, List[str]] = {"command": ["aws"]}
        self.ws_api._on_message(self.mock_ws, message)
        self.assertIsInstance(self.ws_api._error, WebSocketExitCodeError)

    def test_on_message_operation_error(self) -> None:
        """Test handling of operation errors"""
        message: bytes = b"error: something went wrong"
        self.ws_api._on_message(self.mock_ws, message)
        self.assertIsInstance(self.ws_api._error, WebSocketOperationError)

    def test_on_error(self) -> None:
        """Test error handling"""
        error: str = "Connection failed"
        self.ws_api._on_error(self.mock_ws, error)
        self.assertIsInstance(self.ws_api._error, WebSocketConnectionError)
        self.assertIn(error, str(self.ws_api._error))

    def test_on_close_normal(self) -> None:
        """Test normal connection closure"""
        self.ws_api._on_close(self.mock_ws, 1000, "Normal closure")
        self.assertIsNone(self.ws_api._error)

    def test_on_close_unexpected(self) -> None:
        """Test unexpected connection closure"""
        self.ws_api._on_close(self.mock_ws, 1006, "Abnormal closure")
        self.assertIsInstance(self.ws_api._error, WebSocketConnectionError)

    # @patch('websocket.WebSocketApp')
    # def test_exec_success(self, mock_ws_app):
    #     """Test successful execution"""
    #     mock_ws_instance = MagicMock()
    #     mock_ws_app.return_value = mock_ws_instance
    #     request = {"command": "echo hello world"}
    #     result = self.ws_api.exec(**request)
    #     self.assertEqual(result, request)
    #     mock_ws_instance.run_forever.assert_called_once()

    @patch("websocket.WebSocketApp")
    def test_exec_error(self, mock_ws_app: patch) -> None:
        """Test execution with error"""
        mock_ws_instance: MagicMock = MagicMock()
        mock_ws_app.return_value = mock_ws_instance
        self.ws_api._error = WebSocketConnectionError("Test error")

        with self.assertRaises(WebSocketConnectionError):
            self.ws_api.exec(command="test")


if __name__ == "__main__":
    unittest.main()



import requests


class CPLNException(Exception):
    """
    A base class from which all other exceptions inherit.

    If you want to catch all errors that the Docker SDK might raise,
    catch this base exception.
    """


class APIError(requests.exceptions.HTTPError, CPLNException):
    """
    An HTTP error from the API.
    """

    def __init__(self, message, response=None, explanation=None):
        # requests 1.2 supports response as a keyword argument, but
        # requests 1.1 doesn't
        super().__init__(message)
        self.response = response
        self.explanation = explanation

    def __str__(self):
        message = super().__str__()

        if self.is_client_error():
            message = (
                f"{self.response.status_code} Client Error for "
                f"{self.response.url}: {self.response.reason}"
            )

        elif self.is_server_error():
            message = (
                f"{self.response.status_code} Server Error for "
                f"{self.response.url}: {self.response.reason}"
            )

        if self.explanation:
            message = f'{message} ("{self.explanation}")'

        return message

    @property
    def status_code(self):
        if self.response is not None:
            return self.response.status_code

    def is_error(self):
        return self.is_client_error() or self.is_server_error()

    def is_client_error(self):
        if self.status_code is None:
            return False
        return 400 <= self.status_code < 500

    def is_server_error(self):
        if self.status_code is None:
            return False
        return 500 <= self.status_code < 600


class NotFound(APIError):
    pass


class ImageNotFound(NotFound):
    pass


class InvalidVersion(CPLNException):
    pass


class InvalidRepository(CPLNException):
    pass


class InvalidConfigFile(CPLNException):
    pass


class InvalidArgument(CPLNException):
    pass


class DeprecatedMethod(CPLNException):
    pass


class NullResource(CPLNException, ValueError):
    pass


class BuildError(CPLNException):
    def __init__(self, reason, build_log):
        super().__init__(reason)
        self.msg = reason
        self.build_log = build_log


class ImageLoadError(CPLNException):
    pass


class MissingContextParameter(CPLNException):
    def __init__(self, param):
        self.param = param

    def __str__(self):
        return f"missing parameter: {self.param}"


class ContextException(CPLNException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class ContextNotFound(CPLNException):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"context '{self.name}' not found"


class WebSocketError(Exception):
    """Base class for all WebSocket-related errors."""

    pass


class WebSocketConnectionError(WebSocketError):
    """Raised when there are issues with the WebSocket connection."""

    pass


class WebSocketMessageError(WebSocketError):
    """Raised when there are issues with the WebSocket message content."""

    pass


class WebSocketExitCodeError(WebSocketMessageError):
    """Raised when the WebSocket message indicates a non-zero exit code."""

    pass


class WebSocketOperationError(WebSocketMessageError):
    """Raised when the WebSocket message indicates an operation error."""

    pass



import os
from unittest.mock import patch

import pytest
from cpln.constants import (
    DEFAULT_CPLN_API_URL,
)
from cpln.utils import kwargs_from_env

CPLN_TOKEN = os.getenv("CPLN_TOKEN")
CPLN_ORG = os.getenv("CPLN_ORG")


def test_kwargs_from_env_default():
    with patch.dict(os.environ, {"CPLN_TOKEN": CPLN_TOKEN, "CPLN_ORG": CPLN_ORG}):
        kwargs = kwargs_from_env()
        assert kwargs["token"] == CPLN_TOKEN
        assert kwargs["org"] == CPLN_ORG
        assert "base_url" in kwargs
        assert "version" not in kwargs
        assert "timeout" not in kwargs


def test_kwargs_from_env_all_variables():
    with patch.dict(
        os.environ,
        {
            "CPLN_TOKEN": CPLN_TOKEN,
            "CPLN_ORG": CPLN_ORG,
            "CPLN_BASE_URL": DEFAULT_CPLN_API_URL,
            "CPLN_VERSION": "2.0.0",
            "CPLN_TIMEOUT": "30",
        },
    ):
        kwargs = kwargs_from_env()
        assert kwargs["token"] == CPLN_TOKEN
        assert kwargs["org"] == CPLN_ORG
        assert kwargs["base_url"] == DEFAULT_CPLN_API_URL


def test_kwargs_from_env_missing_required():
    with (
        patch.dict(os.environ, {"CPLN_TOKEN": "", "CPLN_ORG": ""}),
        pytest.raises(ValueError),
    ):
        kwargs_from_env()

    with (
        patch.dict(os.environ, {"CPLN_TOKEN": "test-token", "CPLN_ORG": ""}),
        pytest.raises(ValueError),
    ):
        kwargs_from_env()

    with (
        patch.dict(os.environ, {"CPLN_ORG": "test-org", "CPLN_TOKEN": ""}),
        pytest.raises(ValueError),
    ):
        kwargs_from_env()


def test_kwargs_from_env_custom_environment():
    custom_env = {"CPLN_TOKEN": "test-token", "CPLN_ORG": "test-org"}
    kwargs = kwargs_from_env(environment=custom_env)
    assert kwargs["token"] == "test-token"
    assert kwargs["org"] == "test-org"



from .resource import Collection, Model


class Image(Model):
    """
    An image on the server.
    """

    def get(self) -> dict[str, any]:
        """
        Get the image.

        Returns:
            (dict): The image.

        Raises:
            :py:class:`cpln.errors.NotFound`
                If the image does not exist.
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.get_image(self.attrs["name"])

    def delete(self) -> None:
        """
        Delete the image.

        Raises:
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        print(f"Deleting Image: {self}")
        self.client.api.delete_image(self.attrs["name"])
        print("Deleted!")


class ImageCollection(Collection):
    """
    Images on the server.
    """

    model = Image

    def get(self, image_id: str):
        """
        Gets an image.

        Args:
            image_id (str): The name of the image.

        Returns:
            (:py:class:`Image`): The image.

        Raises:
            :py:class:`docker.errors.ImageNotFound`
                If the image does not exist.
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        return self.prepare_model(self.client.api.get_image(image_id))

    def list(self):
        """
        List images on the registry.

        Returns:
            (list of :py:class:`Image`): The images.

        Raises:
            :py:class:`docker.errors.APIError`
                If the server returns an error.
        """
        resp = self.client.api.get_image()["items"]
        return [self.prepare_model(image) for image in resp]



from typing import (
    Any,
)

from ..config import WorkloadConfig
from ..utils import WebSocketAPI
from .config import APIConfig

IGNORED_CONTAINERS = ["cpln-mounter"]


class WorkloadDeploymentMixin:
    """
    A mixin class that provides workload deployment-related API methods.
    """

    def get_workload_deployment(self, config: WorkloadConfig) -> dict[str, Any]:
        """
        Retrieves deployment information for a specific workload.

        Args:
            config (WorkloadConfig): Configuration object containing workload details

        Returns:
            dict: Deployment information for the workload

        Raises:
            ValueError: If the config is not properly set
            APIError: If the request fails
        """
        if config.workload_id is None or config.location is None:
            raise ValueError("Config not set properly")

        endpoint = f"gvc/{config.gvc}/workload/{config.workload_id}/deployment/{config.location}"
        return self._get(endpoint)

    @classmethod
    def get_remote_api(cls, api_config: APIConfig, config: WorkloadConfig):
        """
        Creates a new API client instance for remote operations.

        Args:
            api_config (APIConfig): API configuration
            config (WorkloadConfig): Workload configuration

        Returns:
            WorkloadDeploymentMixin: A new instance configured for remote operations
        """
        env = api_config.asdict()
        env["base_url"] = cls(**env).get_remote(config) + "/replicas"
        return cls(**env)

    def get_remote(self, config: WorkloadConfig) -> str:
        """
        Gets the remote URL for a workload deployment.

        Args:
            config (WorkloadConfig): Configuration object containing workload details

        Returns:
            str: The remote URL for the workload deployment
        """
        return self.get_workload_deployment(config)["status"]["remote"]

    def get_remote_wss(self, config: WorkloadConfig) -> str:
        """
        Gets the WebSocket URL for a workload deployment.

        Args:
            config (WorkloadConfig): Configuration object containing workload details

        Returns:
            str: The WebSocket URL for the workload deployment
        """
        return self.get_remote(config).replace("https:", "wss:") + "/remote"

    def get_replicas(self, config: WorkloadConfig) -> list[str]:
        """
        Gets the list of replicas for a workload.

        Args:
            config (WorkloadConfig): Configuration object containing workload details

        Returns:
            list[str]: List of replica names
        """
        remote_api_client = self.get_remote_api(self.config, config)
        replicas = remote_api_client._get(
            f"/gvc/{config.gvc}/workload/{config.workload_id}"
        )["items"]
        return replicas

    def get_containers(
        self, config: WorkloadConfig, ignored_containers: list[str] = IGNORED_CONTAINERS
    ) -> list[str]:
        """
        Gets the list of containers for a workload, excluding ignored containers.

        Args:
            config (WorkloadConfig): Configuration object containing workload details
            ignored_containers (list[str], optional): List of container names to ignore

        Returns:
            list[str]: List of container names
        """
        item = self.get_workload_deployment(config)
        workload_versions = item["status"]["versions"]
        containers = [
            container
            for ver in workload_versions
            for container, container_specs in ver["containers"].items()
        ]
        return [x for x in containers if x not in ignored_containers]


class WorkloadApiMixin(WorkloadDeploymentMixin):
    """
    A mixin class that provides workload-related API methods.
    """

    def get_workload(self, config: WorkloadConfig):
        """
        Retrieves information about a workload.

        Args:
            config (WorkloadConfig): Configuration object containing workload details

        Returns:
            dict: Workload information

        Raises:
            APIError: If the request fails
        """
        endpoint = f"gvc/{config.gvc}/workload"
        if config.workload_id:
            endpoint += f"/{config.workload_id}"
        return self._get(endpoint)

    def create_workload(
        self,
        config: WorkloadConfig,
        metadata: dict[str, Any],
    ):
        """
        Creates a workload.
        """
        endpoint = f"gvc/{config.gvc}/workload"
        return self._post(endpoint, data=metadata)

    def delete_workload(self, config: WorkloadConfig):
        """
        Deletes a workload.

        Args:
            config (WorkloadConfig): Configuration object containing workload details

        Returns:
            requests.Response: The response from the API

        Raises:
            APIError: If the request fails
        """
        endpoint = f"gvc/{config.gvc}/workload/{config.workload_id}"
        return self._delete(endpoint)

    def patch_workload(self, config: WorkloadConfig, data: dict[str, Any]):
        """
        Updates a workload with the provided data.

        Args:
            config (WorkloadConfig): Configuration object containing workload details
            data (dict): The data to update the workload with

        Returns:
            requests.Response: The response from the API

        Raises:
            APIError: If the request fails
        """
        endpoint = f"gvc/{config.gvc}/workload/{config.workload_id}"
        return self._patch(endpoint, data=data)

    def exec_workload(self, config: WorkloadConfig, command: str):
        """
        Executes a command in a workload container.

        Args:
            config (WorkloadConfig): Configuration object containing workload details
            command (str): The command to execute

        Returns:
            Any: The result of the command execution

        Raises:
            APIError: If the request fails
        """
        containers = self.get_containers(config)
        replicas = self.get_replicas(config)
        remote_wss = self.get_remote_wss(config)
        request = {
            "token": self.config.token,
            "org": self.config.org,
            "gvc": config.gvc,
            "container": containers[-1],
            "pod": replicas[-1],
            "command": command.split(" ") if isinstance(command, str) else command,
        }
        websocket_api = WebSocketAPI(remote_wss)
        return websocket_api.exec(**request)



import unittest
from unittest.mock import MagicMock

from cpln.models.images import Image, ImageCollection


class TestImage(unittest.TestCase):
    def setUp(self):
        self.attrs = {"name": "test-image", "id": "test-image-id"}
        self.client = MagicMock()
        self.collection = MagicMock()
        self.image = Image(
            attrs=self.attrs, client=self.client, collection=self.collection
        )

    def test_get(self):
        """Test get method"""
        expected_response = {"name": "test-image"}
        self.client.api.get_image.return_value = expected_response
        result = self.image.get()
        self.assertEqual(result, expected_response)
        self.client.api.get_image.assert_called_once_with(self.attrs["name"])

    def test_delete(self):
        """Test delete method"""
        self.image.delete()
        self.client.api.delete_image.assert_called_once_with(self.attrs["name"])


class TestImageCollection(unittest.TestCase):
    def setUp(self):
        self.client = MagicMock()
        self.collection = ImageCollection(client=self.client)

    def test_get(self):
        """Test get method"""
        image_id = "test-image"
        expected_image = {"name": image_id}
        self.client.api.get_image.return_value = expected_image
        result = self.collection.get(image_id)
        self.assertIsInstance(result, Image)
        self.assertEqual(result.attrs, expected_image)
        self.client.api.get_image.assert_called_once_with(image_id)

    def test_list(self):
        """Test list method"""
        response = {"items": [{"name": "image1"}, {"name": "image2"}]}
        self.client.api.get_image.return_value = response
        result = self.collection.list()
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], Image)
        self.assertIsInstance(result[1], Image)
        self.assertEqual(result[0].attrs["name"], "image1")
        self.assertEqual(result[1].attrs["name"], "image2")
        self.client.api.get_image.assert_called_once_with()


if __name__ == "__main__":
    unittest.main()



import unittest

from cpln.models.workload_specs import WorkloadAutoscaling, WorkloadSpecState


class TestWorkloadAutoscaling(unittest.TestCase):
    """Tests for the WorkloadAutoscaling class"""

    def setUp(self):
        """Set up the test"""
        self.autoscaling = WorkloadAutoscaling(
            metric="cpu",
            target=80,
            max_scale=10,
            min_scale=1,
            max_concurrency=100,
            scale_to_zero_delay=300,
        )

    def test_initialization(self):
        """Test autoscaling initialization"""
        self.assertEqual(self.autoscaling.metric, "cpu")
        self.assertEqual(self.autoscaling.target, 80)
        self.assertEqual(self.autoscaling.max_scale, 10)
        self.assertEqual(self.autoscaling.min_scale, 1)
        self.assertEqual(self.autoscaling.max_concurrency, 100)
        self.assertEqual(self.autoscaling.scale_to_zero_delay, 300)


class TestWorkloadSpecState(unittest.TestCase):
    """Tests for the WorkloadSpecState class"""

    def setUp(self):
        """Set up the test"""
        self.spec = {
            "defaultOptions": {
                "debug": True,
                "autoscaling": {
                    "metric": "cpu",
                    "target": 80,
                    "maxScale": 10,
                    "minScale": 1,
                    "maxConcurrency": 100,
                    "scaleToZeroDelay": 300,
                },
                "capacityAI": True,
                "suspend": False,
                "timeoutSeconds": 300,
            }
        }

    def test_parse_from_spec(self):
        """Test parse_from_spec class method"""
        spec_state = WorkloadSpecState.parse_from_spec(self.spec)

        # Test basic attributes
        self.assertTrue(spec_state.debug)
        self.assertTrue(spec_state.capacity_ai)
        self.assertFalse(spec_state.suspend)
        self.assertEqual(spec_state.timeout_seconds, 300)

        # Test autoscaling attributes
        self.assertEqual(spec_state.autoscaling.metric, "cpu")
        self.assertEqual(spec_state.autoscaling.target, 80)
        self.assertEqual(spec_state.autoscaling.max_scale, 10)
        self.assertEqual(spec_state.autoscaling.min_scale, 1)
        self.assertEqual(spec_state.autoscaling.max_concurrency, 100)
        self.assertEqual(spec_state.autoscaling.scale_to_zero_delay, 300)

    def test_parse_from_spec_missing_autoscaling(self):
        """Test parse_from_spec with missing autoscaling options"""
        spec = {
            "defaultOptions": {
                "debug": True,
                "capacityAI": True,
                "suspend": False,
                "timeoutSeconds": 300,
            }
        }

        with self.assertRaises(KeyError):
            WorkloadSpecState.parse_from_spec(spec)

    def test_parse_from_spec_missing_default_options(self):
        """Test parse_from_spec with missing defaultOptions"""
        spec = {}

        with self.assertRaises(KeyError):
            WorkloadSpecState.parse_from_spec(spec)

    def test_parse_from_spec_missing_required_fields(self):
        """Test parse_from_spec with missing required fields"""
        spec = {
            "defaultOptions": {
                "autoscaling": {
                    "metric": "cpu",
                    "target": 80,
                    "maxScale": 10,
                    "minScale": 1,
                    "maxConcurrency": 100,
                    "scaleToZeroDelay": 300,
                }
            }
        }

        with self.assertRaises(KeyError):
            WorkloadSpecState.parse_from_spec(spec)


if __name__ == "__main__":
    unittest.main()



from .resource import Collection, Model


class GVC(Model):
    """
    A GVC (Global Virtual Cloud) on the server.
    """

    def get(self) -> dict[str, any]:
        """
        Get the GVC.

        Returns:
            (dict): The GVC.

        Raises:
            :py:class:`cpln.errors.NotFound`
                If the GVC does not exist.
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        return self.client.api.get_gvc(self.attrs["name"])

    def create(self) -> None:
        """
        Create the GVC.

        Raises:
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        print(f"Creating GVC: {self}")
        self.client.api.create_gvc(self.attrs["name"], self.attrs["description"])
        print("Created!")

    def delete(self) -> None:
        """
        Delete the GVC.

        Raises:
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        print(f"Deleting GVC: {self}")
        self.client.api.delete_gvc(self.attrs["name"])
        print("Deleted!")


class GVCCollection(Collection):
    """
    GVCs on the server.
    """

    model = GVC

    def get(self, name: str):
        """
        Get a GVC.

        Args:
            name (str): The name of the GVC.

        Returns:
            (:py:class:`GVC`): The GVC.

        Raises:
            :py:class:`cpln.errors.NotFound`
                If the GVC does not exist.
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        return self.prepare_model(self.client.api.get_gvc(name))

    def list(self):
        """
        List GVCs on the server.

        Returns:
            (list[:py:class:`GVC`]): The GVCs.

        Raises:
            :py:class:`cpln.errors.APIError`
                If the server returns an error.
        """
        resp = self.client.api.get_gvc()["items"]
        return [self.prepare_model(gvc) for gvc in resp]



from dataclasses import dataclass
from typing import TypedDict


class JSON(TypedDict):
    pass


@dataclass
class WorkloadConfig:
    gvc: str
    workload_id: str
    location: str
    specs: JSON



from dataclasses import dataclass

from inflection import underscore


@dataclass
class WorkloadAutoscaling:
    metric: str
    target: int
    max_scale: int
    min_scale: int
    max_concurrency: int
    scale_to_zero_delay: int


@dataclass
class WorkloadSpecState:
    debug: bool
    autoscaling: WorkloadAutoscaling
    capacity_ai: bool
    suspend: bool
    timeout_seconds: int

    @classmethod
    def parse_from_spec(cls, spec: dict) -> "WorkloadSpecState":
        autoscaling = {
            underscore(label): value
            for label, value in spec["defaultOptions"]["autoscaling"].items()
        }
        return cls(
            debug=spec["defaultOptions"]["debug"],
            autoscaling=WorkloadAutoscaling(**autoscaling),
            capacity_ai=spec["defaultOptions"]["capacityAI"],
            suspend=spec["defaultOptions"]["suspend"],
            timeout_seconds=spec["defaultOptions"]["timeoutSeconds"],
        )



import json
import os
from typing import Any

from dotenv import load_dotenv

from ..constants import DEFAULT_CPLN_API_URL

load_dotenv()


def kwargs_from_env(environment=None):
    if not environment:
        environment = os.environ

    base_url = DEFAULT_CPLN_API_URL
    token = environment.get("CPLN_TOKEN")
    org = environment.get("CPLN_ORG")

    params = {}
    if base_url:
        params["base_url"] = base_url

    if token:
        params["token"] = token
    else:
        raise ValueError("CPLN_TOKEN is not set")

    if org:
        params["org"] = org
    else:
        raise ValueError("CPLN_ORG is not set")

    return params


def load_template(template_path: str) -> dict[str, Any]:
    with open(template_path) as file:
        return json.load(file)


def get_default_workload_template(workload_type: str) -> dict[str, Any]:
    if workload_type == "serverless":
        template_path = "../templates/default-serverless-workload.json"
    elif workload_type == "standard":
        template_path = "../templates/default-standard-workload.json"
    else:
        raise ValueError(f"Invalid workload type: {workload_type}")
    spec = load_template(os.path.join(os.path.dirname(__file__), template_path))
    return spec



try:
    from ._version import __version__
except ImportError:
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version("cpln")
    except PackageNotFoundError:
        __version__ = "0.0.0"
